# SPDX-License-Identifier: MIT
import math
import unittest
from datetime import datetime

from odxtools.compumethods.compucategory import CompuCategory
from odxtools.compumethods.compuinternaltophys import CompuInternalToPhys
from odxtools.compumethods.compurationalcoeffs import CompuRationalCoeffs
from odxtools.compumethods.compuscale import CompuScale
from odxtools.compumethods.identicalcompumethod import IdenticalCompuMethod
from odxtools.compumethods.linearcompumethod import LinearCompuMethod
from odxtools.database import Database
from odxtools.dataobjectproperty import DataObjectProperty
from odxtools.decodestate import DecodeState
from odxtools.description import Description
from odxtools.diagdatadictionaryspec import DiagDataDictionarySpec
from odxtools.diaglayers.diaglayertype import DiagLayerType
from odxtools.diaglayers.ecuvariant import EcuVariant
from odxtools.diaglayers.ecuvariantraw import EcuVariantRaw
from odxtools.diagnostictroublecode import DiagnosticTroubleCode
from odxtools.diagservice import DiagService
from odxtools.dtcdop import DtcDop
from odxtools.encodestate import EncodeState
from odxtools.encoding import Encoding
from odxtools.environmentdata import EnvironmentData
from odxtools.environmentdatadescription import EnvironmentDataDescription
from odxtools.exceptions import EncodeError, OdxError
from odxtools.nameditemlist import NamedItemList
from odxtools.odxlink import DocType, OdxDocFragment, OdxLinkDatabase, OdxLinkId, OdxLinkRef
from odxtools.odxtypes import DataType
from odxtools.parameters.codedconstparameter import CodedConstParameter
from odxtools.parameters.nrcconstparameter import NrcConstParameter
from odxtools.parameters.parameter import Parameter
from odxtools.parameters.systemparameter import SystemParameter
from odxtools.parameters.valueparameter import ValueParameter
from odxtools.physicaltype import PhysicalType
from odxtools.radix import Radix
from odxtools.request import Request
from odxtools.response import Response, ResponseType
from odxtools.snrefcontext import SnRefContext
from odxtools.standardlengthtype import StandardLengthType
from odxtools.text import Text

doc_frags = [OdxDocFragment("UnitTest", DocType.CONTAINER)]


class TestEncodeRequest(unittest.TestCase):

    def test_encode_coded_const_infer_order(self) -> None:
        diag_coded_type = StandardLengthType(
            base_data_type=DataType.A_UINT32,
            bit_length=8,
        )
        param1 = CodedConstParameter(
            short_name="coded_const_parameter",
            diag_coded_type=diag_coded_type,
            coded_value_raw=str(0x7D),
            byte_position=0,
        )
        param2 = CodedConstParameter(
            short_name="coded_const_parameter",
            diag_coded_type=diag_coded_type,
            coded_value_raw=str(0xAB),
        )
        req = Request(
            odx_id=OdxLinkId("request_id", doc_frags),
            short_name="request_sn",
            parameters=NamedItemList([param1, param2]),
        )
        self.assertEqual(req.encode(), bytearray([0x7D, 0xAB]))

    def test_string_encodings(self) -> None:
        # ISO 8859-1 (latin1)
        encode_state = EncodeState()
        encode_state.emplace_atomic_value(
            internal_value="ä?a",
            bit_length=24,
            base_data_type=DataType.A_UTF8STRING,
            base_type_encoding=Encoding.ISO_8859_1,
            is_highlow_byte_order=True,
            used_mask=None)
        self.assertEqual(encode_state.coded_message, bytes([0xe4, 0x3f, 0x61]))

        decode_state = DecodeState(encode_state.coded_message)
        decoded = decode_state.extract_atomic_value(
            bit_length=24,
            base_data_type=DataType.A_UTF8STRING,
            base_type_encoding=Encoding.ISO_8859_1,
            is_highlow_byte_order=True)
        self.assertEqual(decoded, "ä?a")

        # UTF-8
        encode_state = EncodeState()
        encode_state.emplace_atomic_value(
            internal_value="ä?",
            bit_length=24,
            base_data_type=DataType.A_UTF8STRING,
            base_type_encoding=Encoding.UTF8,
            is_highlow_byte_order=True,
            used_mask=None)
        self.assertEqual(encode_state.coded_message, bytes([0xc3, 0xa4, 0x3f]))

        decode_state = DecodeState(encode_state.coded_message)
        decoded = decode_state.extract_atomic_value(
            bit_length=24,
            base_data_type=DataType.A_UTF8STRING,
            base_type_encoding=Encoding.UTF8,
            is_highlow_byte_order=True)
        self.assertEqual(decoded, "ä?")

        # big-endian UTF-16
        encode_state = EncodeState()
        encode_state.emplace_atomic_value(
            internal_value="ä",
            bit_length=16,
            base_data_type=DataType.A_UTF8STRING,
            base_type_encoding=Encoding.UCS2,
            is_highlow_byte_order=True,
            used_mask=None)
        self.assertEqual(encode_state.coded_message, bytes([0x00, 0xe4]))

        decode_state = DecodeState(encode_state.coded_message)
        decoded = decode_state.extract_atomic_value(
            bit_length=16,
            base_data_type=DataType.A_UTF8STRING,
            base_type_encoding=Encoding.UCS2,
            is_highlow_byte_order=True)
        self.assertEqual(decoded, "ä")

        # little-endian UTF-16
        encode_state = EncodeState()
        encode_state.emplace_atomic_value(
            internal_value="ä",
            bit_length=16,
            base_data_type=DataType.A_UTF8STRING,
            base_type_encoding=Encoding.UCS2,
            is_highlow_byte_order=False,
            used_mask=None)
        self.assertEqual(encode_state.coded_message, bytes([0xe4, 0x00]))

        decode_state = DecodeState(encode_state.coded_message)
        decoded = decode_state.extract_atomic_value(
            bit_length=16,
            base_data_type=DataType.A_UTF8STRING,
            base_type_encoding=Encoding.UCS2,
            is_highlow_byte_order=False)
        self.assertEqual(decoded, "ä")

    def test_int_encodings(self) -> None:
        # packed BCD
        encode_state = EncodeState()
        encode_state.emplace_atomic_value(
            internal_value=1234,
            bit_length=24,
            base_data_type=DataType.A_UINT32,
            base_type_encoding=Encoding.BCD_P,
            is_highlow_byte_order=True,
            used_mask=None)
        self.assertEqual(encode_state.coded_message, bytes([0x00, 0x12, 0x34]))

        decode_state = DecodeState(encode_state.coded_message)
        decoded = decode_state.extract_atomic_value(
            bit_length=24,
            base_data_type=DataType.A_UINT32,
            base_type_encoding=Encoding.BCD_P,
            is_highlow_byte_order=True)
        self.assertEqual(decoded, 1234)

        # unpacked BCD
        encode_state = EncodeState()
        with self.assertRaises(OdxError):
            encode_state.emplace_atomic_value(
                internal_value=12,
                bit_length=24,
                base_data_type=DataType.A_INT32,
                base_type_encoding=Encoding.BCD_UP,
                is_highlow_byte_order=True,
                used_mask=None)
        encode_state.emplace_atomic_value(
            internal_value=12,
            bit_length=24,
            base_data_type=DataType.A_UINT32,
            base_type_encoding=Encoding.BCD_UP,
            is_highlow_byte_order=True,
            used_mask=None)
        self.assertEqual(encode_state.coded_message, bytes([0x00, 0x01, 0x02]))

        decode_state = DecodeState(encode_state.coded_message)
        decoded = decode_state.extract_atomic_value(
            bit_length=24,
            base_data_type=DataType.A_UINT32,
            base_type_encoding=Encoding.BCD_UP,
            is_highlow_byte_order=True)
        self.assertEqual(decoded, 12)

        # one complement
        encode_state = EncodeState()
        encode_state.emplace_atomic_value(
            internal_value=0x1234,
            bit_length=24,
            base_data_type=DataType.A_INT32,
            base_type_encoding=Encoding.ONEC,
            is_highlow_byte_order=True,
            used_mask=None)
        self.assertEqual(encode_state.coded_message, bytes([0x00, 0x12, 0x34]))

        decode_state = DecodeState(encode_state.coded_message)
        decoded = decode_state.extract_atomic_value(
            bit_length=24,
            base_data_type=DataType.A_INT32,
            base_type_encoding=Encoding.ONEC,
            is_highlow_byte_order=True)
        self.assertEqual(decoded, 0x1234)

        encode_state = EncodeState()
        encode_state.emplace_atomic_value(
            internal_value=-0x1234,
            bit_length=24,
            base_data_type=DataType.A_INT32,
            base_type_encoding=Encoding.ONEC,
            is_highlow_byte_order=True,
            used_mask=None)
        self.assertEqual(encode_state.coded_message, bytes([0xff, 0xed, 0xcb]))

        decode_state = DecodeState(encode_state.coded_message)
        decoded = decode_state.extract_atomic_value(
            bit_length=24,
            base_data_type=DataType.A_INT32,
            base_type_encoding=Encoding.ONEC,
            is_highlow_byte_order=True)
        self.assertEqual(decoded, -0x1234)

        # two complement
        encode_state = EncodeState()
        encode_state.emplace_atomic_value(
            internal_value=0x1234,
            bit_length=24,
            base_data_type=DataType.A_INT32,
            base_type_encoding=Encoding.TWOC,
            is_highlow_byte_order=True,
            used_mask=None)
        self.assertEqual(encode_state.coded_message, bytes([0x00, 0x12, 0x34]))

        decode_state = DecodeState(encode_state.coded_message)
        decoded = decode_state.extract_atomic_value(
            bit_length=24,
            base_data_type=DataType.A_INT32,
            base_type_encoding=Encoding.TWOC,
            is_highlow_byte_order=True)
        self.assertEqual(decoded, 0x1234)

        encode_state = EncodeState()
        encode_state.emplace_atomic_value(
            internal_value=-0x1234,
            bit_length=24,
            base_data_type=DataType.A_INT32,
            base_type_encoding=Encoding.TWOC,
            is_highlow_byte_order=True,
            used_mask=None)
        self.assertEqual(encode_state.coded_message, bytes([0xff, 0xed, 0xcc]))

        decode_state = DecodeState(encode_state.coded_message)
        decoded = decode_state.extract_atomic_value(
            bit_length=24,
            base_data_type=DataType.A_INT32,
            base_type_encoding=Encoding.TWOC,
            is_highlow_byte_order=True)
        self.assertEqual(decoded, -0x1234)

        # sign-magnitude
        encode_state = EncodeState()
        encode_state.emplace_atomic_value(
            internal_value=0x1234,
            bit_length=24,
            base_data_type=DataType.A_INT32,
            base_type_encoding=Encoding.SM,
            is_highlow_byte_order=True,
            used_mask=None)
        self.assertEqual(encode_state.coded_message, bytes([0x00, 0x12, 0x34]))

        decode_state = DecodeState(encode_state.coded_message)
        decoded = decode_state.extract_atomic_value(
            bit_length=24,
            base_data_type=DataType.A_INT32,
            base_type_encoding=Encoding.SM,
            is_highlow_byte_order=True)
        self.assertEqual(decoded, 0x1234)

        encode_state = EncodeState()
        encode_state.emplace_atomic_value(
            internal_value=-0x1234,
            bit_length=24,
            base_data_type=DataType.A_INT32,
            base_type_encoding=Encoding.SM,
            is_highlow_byte_order=True,
            used_mask=None)
        self.assertEqual(encode_state.coded_message, bytes([0x80, 0x12, 0x34]))

        decode_state = DecodeState(encode_state.coded_message)
        decoded = decode_state.extract_atomic_value(
            bit_length=24,
            base_data_type=DataType.A_INT32,
            base_type_encoding=Encoding.SM,
            is_highlow_byte_order=True)
        self.assertEqual(decoded, -0x1234)

    def test_float_encodings(self) -> None:
        # FLOAT32
        encode_state = EncodeState()
        with self.assertRaises(OdxError):
            encode_state.emplace_atomic_value(
                internal_value=-1.234,
                bit_length=24,
                base_data_type=DataType.A_FLOAT32,
                base_type_encoding=Encoding.NONE,
                is_highlow_byte_order=True,
                used_mask=None)
        encode_state.emplace_atomic_value(
            internal_value=-1.234,
            bit_length=32,
            base_data_type=DataType.A_FLOAT32,
            base_type_encoding=Encoding.NONE,
            is_highlow_byte_order=True,
            used_mask=None)
        self.assertEqual(encode_state.coded_message, b'\xbf\x9d\xf3\xb6')

        decode_state = DecodeState(encode_state.coded_message)
        with self.assertRaises(OdxError):
            decoded = decode_state.extract_atomic_value(
                bit_length=24,
                base_data_type=DataType.A_FLOAT32,
                base_type_encoding=Encoding.NONE,
                is_highlow_byte_order=True)
        decoded = decode_state.extract_atomic_value(
            bit_length=32,
            base_data_type=DataType.A_FLOAT32,
            base_type_encoding=Encoding.NONE,
            is_highlow_byte_order=True)
        # allow rounding errors due to python's float objects
        # potentially using a different representation
        assert isinstance(decoded, float)
        self.assertTrue(abs(decoded - (-1.234)) < 1e-6)

        # FLOAT32 little-endian
        encode_state = EncodeState()
        encode_state.emplace_atomic_value(
            internal_value=-1.234,
            bit_length=32,
            base_data_type=DataType.A_FLOAT32,
            base_type_encoding=Encoding.NONE,
            is_highlow_byte_order=False,
            used_mask=None)
        self.assertEqual(encode_state.coded_message, b'\xb6\xf3\x9d\xbf')

        decode_state = DecodeState(encode_state.coded_message)
        decoded = decode_state.extract_atomic_value(
            bit_length=32,
            base_data_type=DataType.A_FLOAT32,
            base_type_encoding=Encoding.NONE,
            is_highlow_byte_order=False)
        # allow rounding errors due to python's float objects
        # potentially using a different representation
        assert isinstance(decoded, float)
        self.assertTrue(abs(decoded - (-1.234)) < 1e-6)

        # check if NaN can be handled
        encode_state = EncodeState()
        encode_state.emplace_atomic_value(
            internal_value=float("nan"),
            bit_length=32,
            base_data_type=DataType.A_FLOAT32,
            base_type_encoding=Encoding.NONE,
            is_highlow_byte_order=True,
            used_mask=None)
        self.assertEqual(encode_state.coded_message, b'\x7f\xc0\x00\x00')

        decode_state = DecodeState(encode_state.coded_message)
        decoded = decode_state.extract_atomic_value(
            bit_length=32,
            base_data_type=DataType.A_FLOAT32,
            base_type_encoding=Encoding.NONE,
            is_highlow_byte_order=True)
        assert isinstance(decoded, float)
        self.assertTrue(math.isnan(decoded))

        # FLOAT64
        encode_state = EncodeState()
        with self.assertRaises(OdxError):
            encode_state.emplace_atomic_value(
                internal_value=-1.234,
                bit_length=24,
                base_data_type=DataType.A_FLOAT64,
                base_type_encoding=Encoding.NONE,
                is_highlow_byte_order=True,
                used_mask=None)
        encode_state.emplace_atomic_value(
            internal_value=-1.234,
            bit_length=64,
            base_data_type=DataType.A_FLOAT64,
            base_type_encoding=Encoding.NONE,
            is_highlow_byte_order=True,
            used_mask=None)
        self.assertEqual(encode_state.coded_message, b'\xbf\xf3\xbev\xc8\xb49X')

        decode_state = DecodeState(encode_state.coded_message)
        with self.assertRaises(OdxError):
            decoded = decode_state.extract_atomic_value(
                bit_length=24,
                base_data_type=DataType.A_FLOAT64,
                base_type_encoding=Encoding.NONE,
                is_highlow_byte_order=True)
        decoded = decode_state.extract_atomic_value(
            bit_length=64,
            base_data_type=DataType.A_FLOAT64,
            base_type_encoding=Encoding.NONE,
            is_highlow_byte_order=True)
        # allow rounding errors due to python's float objects
        # potentially using a different representation
        assert isinstance(decoded, float)
        self.assertTrue(abs(decoded - (-1.234)) < 1e-9)

    def test_encode_coded_const_reorder(self) -> None:
        diag_coded_type = StandardLengthType(
            base_data_type=DataType.A_UINT32,
            bit_length=8,
        )
        param1 = CodedConstParameter(
            short_name="param1",
            diag_coded_type=diag_coded_type,
            coded_value_raw=str(0x34),
            byte_position=1,
        )
        param2 = CodedConstParameter(
            short_name="param2",
            diag_coded_type=diag_coded_type,
            coded_value_raw=str(0x12),
            byte_position=0,
        )
        req = Request(
            odx_id=OdxLinkId("request_id", doc_frags),
            short_name="request_sn",
            parameters=NamedItemList([param1, param2]),
        )
        self.assertEqual(req.encode(), bytearray([0x12, 0x34]))

    def test_encode_linear(self) -> None:
        odxlinks = OdxLinkDatabase()
        diag_coded_type = StandardLengthType(
            base_data_type=DataType.A_UINT32,
            bit_length=8,
        )
        # This CompuMethod represents the linear function: decode(x) = 2*x + 8 and encode(x) = (x-8)/2
        compu_method = LinearCompuMethod(
            category=CompuCategory.LINEAR,
            compu_internal_to_phys=CompuInternalToPhys(compu_scales=[
                CompuScale(
                    compu_rational_coeffs=CompuRationalCoeffs(
                        value_type=DataType.A_INT32,
                        numerators=[8, 2],
                        denominators=[1],
                    ),
                    domain_type=DataType.A_INT32,
                    range_type=DataType.A_INT32),
            ]),
            internal_type=DataType.A_UINT32,
            physical_type=DataType.A_UINT32)
        dop = DataObjectProperty(
            odx_id=OdxLinkId("dop.id", doc_frags),
            short_name="dop_sn",
            long_name="example dop",
            diag_coded_type=diag_coded_type,
            physical_type=PhysicalType(base_data_type=DataType.A_UINT32),
            compu_method=compu_method,
        )
        odxlinks.update({dop.odx_id: dop})
        param1 = ValueParameter(
            short_name="value_parameter",
            dop_ref=OdxLinkRef.from_id(dop.odx_id),
        )
        req = Request(
            odx_id=OdxLinkId("request.id", doc_frags),
            short_name="request_sn",
            parameters=NamedItemList([param1]),
        )

        snref_ctx = SnRefContext(parameters=req.parameters)
        param1._resolve_odxlinks(odxlinks)
        param1._resolve_snrefs(snref_ctx)

        # Missing mandatory parameter.
        with self.assertRaises(EncodeError):
            req.encode()

        self.assertEqual(
            req.encode(value_parameter=14),
            bytearray([0x3])  # encode(14) = (14-8)/2 = 3
        )

    def test_encode_nrc_const(self) -> None:
        diag_coded_type = StandardLengthType(
            base_data_type=DataType.A_UINT32,
            bit_length=8,
        )
        dop = DataObjectProperty(
            odx_id=OdxLinkId("dop.id", doc_frags),
            short_name="dop_sn",
            long_name="example dop",
            diag_coded_type=diag_coded_type,
            physical_type=PhysicalType(base_data_type=DataType.A_UINT32),
            compu_method=IdenticalCompuMethod(
                category=CompuCategory.IDENTICAL,
                internal_type=DataType.A_UINT32,
                physical_type=DataType.A_UINT32),
        )
        param1 = CodedConstParameter(
            short_name="param1",
            diag_coded_type=diag_coded_type,
            coded_value_raw=str(0x12),
            byte_position=0,
        )
        param2 = NrcConstParameter(
            short_name="param2",
            diag_coded_type=diag_coded_type,
            coded_values_raw=[str(0x34), str(0xAB)],
            byte_position=1,
        )
        param3 = ValueParameter(
            short_name="param3",
            dop_ref=OdxLinkRef.from_id(dop.odx_id),
            byte_position=1,
        )
        resp = Response(
            odx_id=OdxLinkId("response_id", doc_frags),
            short_name="response_sn",
            response_type=ResponseType.POSITIVE,
            parameters=NamedItemList([param1, param2, param3]),
        )

        req = Request(
            odx_id=OdxLinkId("request_id", doc_frags),
            short_name="request_sn",
            parameters=NamedItemList([
                CodedConstParameter(
                    short_name="req_param1",
                    diag_coded_type=diag_coded_type,
                    coded_value_raw=str(0xB0),
                    byte_position=0,
                )
            ]),
        )

        service = DiagService(
            odx_id=OdxLinkId("service_id", doc_frags),
            short_name="service_sn",
            request_ref=OdxLinkRef.from_id(req.odx_id),
            neg_response_refs=[OdxLinkRef.from_id(resp.odx_id)],
        )

        ecu_variant_raw = EcuVariantRaw(
            variant_type=DiagLayerType.ECU_VARIANT,
            odx_id=OdxLinkId("dl_id", doc_frags),
            short_name="dl_sn",
            diag_data_dictionary_spec=DiagDataDictionarySpec(
                data_object_props=NamedItemList([dop])),
            diag_comms_raw=[service],
            requests=NamedItemList([req]),
            negative_responses=NamedItemList([resp]),
        )

        ecu_variant = EcuVariant(diag_layer_raw=ecu_variant_raw)
        db = Database()
        odxlinks = OdxLinkDatabase()
        odxlinks.update(ecu_variant._build_odxlinks())
        ecu_variant._resolve_odxlinks(odxlinks)
        ecu_variant._finalize_init(db, odxlinks)

        with self.assertRaises(EncodeError):
            resp.encode()  # "No value for required parameter param3 specified"
        self.assertEqual(resp.encode(param3=0xAB).hex(), "12ab")
        with self.assertRaises(EncodeError):
            # Should raise an EncodeError because the value of
            # NRC-CONST parameters cannot be directly specified
            resp.encode(param2=0xEF, param3=0xAB)

    def test_encode_system_parameter(self) -> None:
        diag_coded_type = StandardLengthType(
            base_data_type=DataType.A_UINT32,
            bit_length=16,
        )
        dop = DataObjectProperty(
            odx_id=OdxLinkId("dop.year", doc_frags),
            short_name="dop_year_sn",
            diag_coded_type=diag_coded_type,
            physical_type=PhysicalType(base_data_type=DataType.A_UINT32),
            compu_method=IdenticalCompuMethod(
                category=CompuCategory.IDENTICAL,
                internal_type=DataType.A_UINT32,
                physical_type=DataType.A_UINT32),
        )
        param1 = SystemParameter(
            short_name="year_param",
            dop_ref=OdxLinkRef.from_id(dop.odx_id),
            sysparam="YEAR",
            byte_position=0,
        )

        odxlinks = OdxLinkDatabase()
        odxlinks.update(dop._build_odxlinks())

        param1._resolve_odxlinks(odxlinks)

        encode_state = EncodeState()
        param1.encode_into_pdu(physical_value=2024, encode_state=encode_state)
        self.assertEqual(encode_state.coded_message, b'\x07\xe8')

        # test auto-determination of parameter value
        cur_year = datetime.now().year
        encode_state = EncodeState()
        param1.encode_into_pdu(physical_value=None, encode_state=encode_state)

        # there is a (rather theoretical) race condition here: if the
        # cur_year variable was assigned before the year of the system
        # date changes (e.g., because it is new-year's eve) and the
        # encoding was done after that, we will get an incorrect value
        # here. (good luck exploiting this!)
        self.assertEqual(encode_state.coded_message, cur_year.to_bytes(2, 'big'))

        # ensure that decoding works as well
        decode_state = DecodeState(coded_message=encode_state.coded_message)
        self.assertEqual(param1.decode_from_pdu(decode_state), cur_year)

    def test_encode_env_data_desc(self) -> None:
        dct = StandardLengthType(
            base_data_type=DataType.A_UINT32,
            bit_length=8,
        )
        dop = DataObjectProperty(
            odx_id=OdxLinkId("dop.id", doc_frags),
            short_name="dop_sn",
            long_name="example dop",
            diag_coded_type=dct,
            physical_type=PhysicalType(base_data_type=DataType.A_UINT32),
            compu_method=IdenticalCompuMethod(
                category=CompuCategory.IDENTICAL,
                internal_type=DataType.A_UINT32,
                physical_type=DataType.A_UINT32),
        )

        dtc_dct = StandardLengthType(
            base_data_type=DataType.A_UINT32,
            bit_length=24,
        )
        dtc_dop = DtcDop(
            odx_id=OdxLinkId("dtcdop.id", doc_frags),
            short_name="dtcdop_sn",
            description=Description.from_string(
                "DOP containing all possible diagnostic trouble codes"),
            diag_coded_type=dtc_dct,
            physical_type=PhysicalType(base_data_type=DataType.A_UINT32, display_radix=Radix.HEX),
            compu_method=IdenticalCompuMethod(
                category=CompuCategory.IDENTICAL,
                internal_type=DataType.A_UINT32,
                physical_type=DataType.A_UINT32),
            dtcs_raw=[
                DiagnosticTroubleCode(
                    odx_id=OdxLinkId("DTCs.first_trouble", doc_frags),
                    short_name="first_trouble",
                    trouble_code=0x112233,
                    text=Text.from_string("The first trouble is the deepest"),
                    display_trouble_code="Z123"),
                DiagnosticTroubleCode(
                    odx_id=OdxLinkId("DTCs.follow_up_trouble", doc_frags),
                    short_name="follow_up_trouble",
                    trouble_code=0x445566,
                    text=Text.from_string(""),
                    display_trouble_code="Y456"),
                DiagnosticTroubleCode(
                    odx_id=OdxLinkId("DTCs.screwed_up_hard", doc_frags),
                    short_name="screwed_up_hard",
                    trouble_code=0xf00de5,
                    text=Text.from_string(""),
                    display_trouble_code="SCREW"),
            ],
        )

        env_data_desc = EnvironmentDataDescription(
            odx_id=OdxLinkId("DTCs.trouble_explanation", doc_frags),
            short_name="trouble_explanation",
            param_snref="DTC",
            env_datas=NamedItemList([
                EnvironmentData(
                    odx_id=OdxLinkId("DTCs.trouble_explanation.boiler_plate", doc_frags),
                    short_name="boiler_plate",
                    all_value=True,
                    parameters=NamedItemList([
                        CodedConstParameter(
                            short_name="blabla_boiler",
                            diag_coded_type=dct,
                            coded_value_raw=str(0xee),
                            byte_position=0,
                        )
                    ])),
                EnvironmentData(
                    odx_id=OdxLinkId("DTCs.trouble_explanation.reason_for_1", doc_frags),
                    short_name="reason_for_1",
                    dtc_values=[0x112233],
                    parameters=NamedItemList([
                        CodedConstParameter(
                            short_name="blabla_1",
                            diag_coded_type=dct,
                            coded_value_raw=str(0x01),
                        )
                    ])),
                EnvironmentData(
                    odx_id=OdxLinkId("DTCs.trouble_explanation.reason_for_2", doc_frags),
                    short_name="reason_for_2",
                    dtc_values=[0x445566],
                    parameters=NamedItemList([
                        CodedConstParameter(
                            short_name="blabla_3",
                            diag_coded_type=dct,
                            coded_value_raw=str(0x03),
                            byte_position=1,
                        ),
                        CodedConstParameter(
                            short_name="blabla_2",
                            diag_coded_type=dct,
                            coded_value_raw=str(0x02),
                            byte_position=0,
                        ),
                    ])),
            ]),
        )

        param1 = ValueParameter(
            short_name="DTC",
            dop_ref=OdxLinkRef.from_id(dtc_dop.odx_id),
        )
        param2 = ValueParameter(
            short_name="dtc_info",
            description=Description.from_string("Supplemental info why the error happened"),
            dop_ref=OdxLinkRef.from_id(env_data_desc.odx_id),
        )

        resp = Response(
            odx_id=OdxLinkId("DTCs.report_dtc.answer", doc_frags),
            short_name="report_dtc_answer",
            parameters=NamedItemList([param1, param2]),
            response_type=ResponseType.POSITIVE,
        )

        odxlinks = OdxLinkDatabase()
        odxlinks.update(dop._build_odxlinks())
        odxlinks.update(dtc_dop._build_odxlinks())
        odxlinks.update(env_data_desc._build_odxlinks())
        odxlinks.update(resp._build_odxlinks())

        dop._resolve_odxlinks(odxlinks)
        dtc_dop._resolve_odxlinks(odxlinks)
        env_data_desc._resolve_odxlinks(odxlinks)
        resp._resolve_odxlinks(odxlinks)

        snref_ctx = SnRefContext()

        dop._resolve_snrefs(snref_ctx)
        dtc_dop._resolve_snrefs(snref_ctx)
        env_data_desc._resolve_snrefs(snref_ctx)
        resp._resolve_snrefs(snref_ctx)

        # test environment data for DCT 0x112233
        raw_data = resp.encode(DTC=0x112233, dtc_info={})
        self.assertEqual(raw_data.hex(), "112233ee01")

        # test environment data for DCT 0x445566
        raw_data = resp.encode(DTC=0x445566, dtc_info={})
        self.assertEqual(raw_data.hex(), "445566ee0203")

        # test for a DCT without any special environment data (just
        # the all-data boiler plate)
        raw_data = resp.encode(DTC=0xf00de5, dtc_info={})
        self.assertEqual(raw_data.hex(), "f00de5ee")

        # test an unspecified DCT (raises EncodeError)
        with self.assertRaises(EncodeError):
            raw_data = resp.encode(DTC=0x00c007, dtc_info={})

    def test_encode_overlapping(self) -> None:
        uint24 = StandardLengthType(
            base_data_type=DataType.A_UINT32,
            bit_length=24,
        )
        uint8 = StandardLengthType(
            base_data_type=DataType.A_UINT32,
            bit_length=8,
        )
        param1 = CodedConstParameter(
            short_name="code",
            diag_coded_type=uint24,
            coded_value_raw=str(0x123456),
            byte_position=0,
        )
        param2 = CodedConstParameter(
            short_name="part1",
            diag_coded_type=uint8,
            coded_value_raw=str(0x23),
            byte_position=0,
            bit_position=4,
        )
        param3 = CodedConstParameter(
            short_name="part2",
            diag_coded_type=uint8,
            coded_value_raw=str(0x45),
            byte_position=1,
            bit_position=4,
        )
        req = Request(
            odx_id=OdxLinkId("request_id", doc_frags),
            short_name="request_sn",
            parameters=NamedItemList([param1, param2, param3]),
        )
        self.assertEqual(req.encode().hex(), "123456")
        self.assertEqual(req.get_static_bit_length(), 24)

    def _create_request(self, parameters: list[Parameter]) -> Request:
        return Request(
            odx_id=OdxLinkId("request_id", doc_frags),
            short_name="request_sn",
            parameters=NamedItemList(parameters),
        )

    def test_bit_mask(self) -> None:
        inner_dct = StandardLengthType(
            bit_mask=0x3fc, base_data_type=DataType.A_UINT32, bit_length=14)
        outer_dct = StandardLengthType(
            bit_mask=0xf00f, base_data_type=DataType.A_UINT32, bit_length=16)

        physical_type = PhysicalType(base_data_type=DataType.A_UINT32)
        compu_method = IdenticalCompuMethod(
            category=CompuCategory.IDENTICAL,
            internal_type=DataType.A_UINT32,
            physical_type=DataType.A_UINT32)

        inner_dop = DataObjectProperty(
            odx_id=OdxLinkId('dop.inner', doc_frags),
            short_name="inner_dop",
            diag_coded_type=inner_dct,
            physical_type=physical_type,
            compu_method=compu_method)

        outer_dop = DataObjectProperty(
            odx_id=OdxLinkId('dop.outer', doc_frags),
            short_name="outer_dop",
            diag_coded_type=outer_dct,
            physical_type=physical_type,
            compu_method=compu_method)

        odxlinks = OdxLinkDatabase()
        odxlinks.update(inner_dop._build_odxlinks())
        odxlinks.update(outer_dop._build_odxlinks())
        odxlinks.update(inner_dct._build_odxlinks())
        odxlinks.update(outer_dct._build_odxlinks())

        # Inner
        inner_param = ValueParameter(
            short_name="inner_param",
            byte_position=0,
            bit_position=2,
            dop_ref=OdxLinkRef.from_id(inner_dop.odx_id))
        snref_ctx = SnRefContext()
        inner_param._resolve_odxlinks(odxlinks)
        inner_param._resolve_snrefs(snref_ctx)

        # Outer
        outer_param = ValueParameter(
            short_name="outer_param", byte_position=0, dop_ref=OdxLinkRef.from_id(outer_dop.odx_id))
        outer_param._resolve_odxlinks(odxlinks)
        outer_param._resolve_snrefs(snref_ctx)

        req = self._create_request([inner_param, outer_param])

        # the bit shifts here stem from the fact that we placed the
        # inner parameter at bit position 2...
        self.assertEqual(req.encode(inner_param=0x1234 >> 2, outer_param=0x4568).hex(), "4238")
        self.assertEqual(
            req.decode(bytes.fromhex('abcd')), {
                "inner_param": (0xbc << 2),
                "outer_param": 0xa00d
            })

    def test_condensed_bit_mask(self) -> None:
        dct = StandardLengthType(
            bit_mask=0xf00f, base_data_type=DataType.A_UINT32, is_condensed_raw=True, bit_length=16)

        self.assertEqual(dct.get_static_bit_length(), 8)

        physical_type = PhysicalType(base_data_type=DataType.A_UINT32)
        compu_method = IdenticalCompuMethod(
            category=CompuCategory.IDENTICAL,
            internal_type=DataType.A_UINT32,
            physical_type=DataType.A_UINT32)

        dop = DataObjectProperty(
            odx_id=OdxLinkId('dop.inner', doc_frags),
            short_name="dop",
            diag_coded_type=dct,
            physical_type=physical_type,
            compu_method=compu_method)

        odxlinks = OdxLinkDatabase()
        odxlinks.update(dop._build_odxlinks())
        odxlinks.update(dct._build_odxlinks())

        # Inner
        param = ValueParameter(
            short_name="param",
            byte_position=0,
            bit_position=2,
            dop_ref=OdxLinkRef.from_id(dop.odx_id))
        param._resolve_odxlinks(odxlinks)

        req = self._create_request([param])

        # the values here stem from the fact that we placed the
        # parameter at bit position 2...
        self.assertEqual(req.encode(param=0x1ff4).hex(), "000050")
        self.assertEqual(req.decode(bytes.fromhex('000050')), {
            "param": 0x1004,
        })


if __name__ == "__main__":
    unittest.main()



# SPDX-License-Identifier: MIT
from dataclasses import dataclass
from typing import Any
from xml.etree import ElementTree

from .element import IdentifiableElement
from .odxdoccontext import OdxDocContext
from .odxlink import OdxLinkDatabase, OdxLinkId
from .snrefcontext import SnRefContext
from .utils import dataclass_fields_asdict


@dataclass(kw_only=True)
class AdditionalAudience(IdentifiableElement):
    """
    Corresponds to ADDITIONAL-AUDIENCE.
    """

    @staticmethod
    def from_et(et_element: ElementTree.Element, context: OdxDocContext) -> "AdditionalAudience":
        kwargs = dataclass_fields_asdict(IdentifiableElement.from_et(et_element, context))

        return AdditionalAudience(**kwargs)

    def _build_odxlinks(self) -> dict[OdxLinkId, Any]:
        return {self.odx_id: self}

    def _resolve_odxlinks(self, odxlinks: OdxLinkDatabase) -> None:
        pass

    def _resolve_snrefs(self, context: SnRefContext) -> None:
        pass



# SPDX-License-Identifier: MIT
from dataclasses import dataclass
from xml.etree import ElementTree

from .element import IdentifiableElement
from .exceptions import odxrequire
from .odxdoccontext import OdxDocContext
from .utils import dataclass_fields_asdict


@dataclass(kw_only=True)
class ExternalAccessMethod(IdentifiableElement):
    method: str

    @staticmethod
    def from_et(et_element: ElementTree.Element, context: OdxDocContext) -> "ExternalAccessMethod":
        kwargs = dataclass_fields_asdict(IdentifiableElement.from_et(et_element, context))

        method = odxrequire(et_element.findtext("METHOD"))

        return ExternalAccessMethod(method=method, **kwargs)



# SPDX-License-Identifier: MIT
from dataclasses import dataclass
from xml.etree import ElementTree

from .basicstructure import BasicStructure
from .odxdoccontext import OdxDocContext
from .odxtypes import odxstr_to_bool
from .utils import dataclass_fields_asdict


@dataclass(kw_only=True)
class Structure(BasicStructure):
    is_visible_raw: bool | None = None

    @property
    def is_visible(self) -> bool:
        return self.is_visible_raw in (True, None)

    @staticmethod
    def from_et(et_element: ElementTree.Element, context: OdxDocContext) -> "Structure":
        """Read a STRUCTURE element from XML."""
        kwargs = dataclass_fields_asdict(BasicStructure.from_et(et_element, context))

        is_visible_raw = odxstr_to_bool(et_element.get("IS-VISIBLE"))

        return Structure(is_visible_raw=is_visible_raw, **kwargs)



# SPDX-License-Identifier: MIT
from dataclasses import dataclass, field
from typing import Any
from xml.etree import ElementTree

from .admindata import AdminData
from .dataobjectproperty import DataObjectProperty
from .element import IdentifiableElement
from .exceptions import odxassert
from .nameditemlist import NamedItemList
from .odxdoccontext import OdxDocContext
from .odxlink import OdxLinkDatabase, OdxLinkId, OdxLinkRef
from .snrefcontext import SnRefContext
from .specialdatagroup import SpecialDataGroup
from .tablediagcommconnector import TableDiagCommConnector
from .tablerow import TableRow
from .utils import dataclass_fields_asdict


@dataclass(kw_only=True)
class Table(IdentifiableElement):
    """This class represents a TABLE."""
    key_label: str | None = None
    struct_label: str | None = None
    admin_data: AdminData | None = None
    key_dop_ref: OdxLinkRef | None = None
    table_rows_raw: list[TableRow | OdxLinkRef] = field(default_factory=list)
    table_diag_comm_connectors: list[TableDiagCommConnector] = field(default_factory=list)
    sdgs: list[SpecialDataGroup] = field(default_factory=list)
    semantic: str | None = None

    @property
    def key_dop(self) -> DataObjectProperty | None:
        """The key data object property associated with this table."""
        return self._key_dop

    @property
    def table_rows(self) -> NamedItemList[TableRow]:
        """The table rows (both local and referenced) in this table."""
        return self._table_rows

    @staticmethod
    def from_et(et_element: ElementTree.Element, context: OdxDocContext) -> "Table":
        """Reads a TABLE."""
        kwargs = dataclass_fields_asdict(IdentifiableElement.from_et(et_element, context))
        odx_id = kwargs["odx_id"]
        key_label = et_element.findtext("KEY-LABEL")
        struct_label = et_element.findtext("STRUCT-LABEL")
        admin_data = AdminData.from_et(et_element.find("ADMIN-DATA"), context)
        key_dop_ref = OdxLinkRef.from_et(et_element.find("KEY-DOP-REF"), context)

        table_rows_raw: list[OdxLinkRef | TableRow] = []
        for sub_elem in et_element:
            if sub_elem.tag == "TABLE-ROW":
                table_rows_raw.append(
                    TableRow.tablerow_from_et(
                        sub_elem, context, table_ref=OdxLinkRef.from_id(odx_id)))
            elif sub_elem.tag == "TABLE-ROW-REF":
                table_rows_raw.append(OdxLinkRef.from_et(sub_elem, context))

        table_diag_comm_connectors = [
            TableDiagCommConnector.from_et(dcc_elem, context) for dcc_elem in et_element.iterfind(
                "TABLE-DIAG-COMM-CONNECTORS/TABLE-DIAG-COMM-CONNECTOR")
        ]
        sdgs = [SpecialDataGroup.from_et(sdge, context) for sdge in et_element.iterfind("SDGS/SDG")]
        semantic = et_element.get("SEMANTIC")

        return Table(
            key_label=key_label,
            struct_label=struct_label,
            admin_data=admin_data,
            key_dop_ref=key_dop_ref,
            table_rows_raw=table_rows_raw,
            table_diag_comm_connectors=table_diag_comm_connectors,
            sdgs=sdgs,
            semantic=semantic,
            **kwargs)

    def _build_odxlinks(self) -> dict[OdxLinkId, Any]:
        result = {self.odx_id: self}

        for table_row_wrapper in self.table_rows_raw:
            if isinstance(table_row_wrapper, TableRow):
                result.update(table_row_wrapper._build_odxlinks())

        for dcc in self.table_diag_comm_connectors:
            result.update(dcc._build_odxlinks())

        for sdg in self.sdgs:
            result.update(sdg._build_odxlinks())

        return result

    def _resolve_odxlinks(self, odxlinks: OdxLinkDatabase) -> None:
        self._key_dop: DataObjectProperty | None = None
        if self.key_dop_ref is not None:
            self._key_dop = odxlinks.resolve(self.key_dop_ref, DataObjectProperty)

        table_rows = []
        for table_row_wrapper in self.table_rows_raw:
            if isinstance(table_row_wrapper, TableRow):
                table_row = table_row_wrapper
                table_row._resolve_odxlinks(odxlinks)
            else:
                odxassert(isinstance(table_row_wrapper, OdxLinkRef))
                table_row = odxlinks.resolve(table_row_wrapper, TableRow)

            table_rows.append(table_row)

        self._table_rows = NamedItemList(table_rows)

        for dcc in self.table_diag_comm_connectors:
            dcc._resolve_odxlinks(odxlinks)

        for sdg in self.sdgs:
            sdg._resolve_odxlinks(odxlinks)

    def _resolve_snrefs(self, context: SnRefContext) -> None:
        for table_row_wrapper in self.table_rows_raw:
            if isinstance(table_row_wrapper, TableRow):
                table_row_wrapper._resolve_snrefs(context)

        for dcc in self.table_diag_comm_connectors:
            dcc._resolve_snrefs(context)

        for sdg in self.sdgs:
            sdg._resolve_snrefs(context)



# SPDX-License-Identifier: MIT
from dataclasses import dataclass
from xml.etree import ElementTree

from .exceptions import odxrequire
from .odxdoccontext import OdxDocContext
from .odxlink import OdxLinkRef
from .utils import dataclass_fields_asdict


@dataclass(kw_only=True)
class RelatedDiagCommRef(OdxLinkRef):
    relation_type: str

    @staticmethod
    def from_et(  # type: ignore[override]
            et_element: ElementTree.Element, context: OdxDocContext) -> "RelatedDiagCommRef":
        kwargs = dataclass_fields_asdict(odxrequire(OdxLinkRef.from_et(et_element, context)))

        relation_type = odxrequire(et_element.findtext("RELATION-TYPE"))

        return RelatedDiagCommRef(relation_type=relation_type, **kwargs)



# SPDX-License-Identifier: MIT
import pickle
import unittest

from odxtools.compumethods.compucategory import CompuCategory
from odxtools.compumethods.compuinternaltophys import CompuInternalToPhys
from odxtools.compumethods.compurationalcoeffs import CompuRationalCoeffs
from odxtools.compumethods.compuscale import CompuScale
from odxtools.compumethods.identicalcompumethod import IdenticalCompuMethod
from odxtools.compumethods.linearcompumethod import LinearCompuMethod
from odxtools.database import Database
from odxtools.dataobjectproperty import DataObjectProperty
from odxtools.determinenumberofitems import DetermineNumberOfItems
from odxtools.diagdatadictionaryspec import DiagDataDictionarySpec
from odxtools.diaglayers.basevariant import BaseVariant
from odxtools.diaglayers.basevariantraw import BaseVariantRaw
from odxtools.diaglayers.diaglayertype import DiagLayerType
from odxtools.diaglayers.ecuvariant import EcuVariant
from odxtools.diaglayers.ecuvariantraw import EcuVariantRaw
from odxtools.diagnostictroublecode import DiagnosticTroubleCode
from odxtools.diagservice import DiagService
from odxtools.dtcdop import DtcDop
from odxtools.dynamicendmarkerfield import DynamicEndmarkerField
from odxtools.dynamiclengthfield import DynamicLengthField
from odxtools.dynenddopref import DynEndDopRef
from odxtools.endofpdufield import EndOfPduField
from odxtools.exceptions import DecodeError, DecodeMismatch
from odxtools.message import Message
from odxtools.minmaxlengthtype import MinMaxLengthType
from odxtools.nameditemlist import NamedItemList
from odxtools.odxlink import DocType, OdxDocFragment, OdxLinkDatabase, OdxLinkId, OdxLinkRef
from odxtools.odxtypes import DataType, ParameterValueDict
from odxtools.parameters.codedconstparameter import CodedConstParameter
from odxtools.parameters.matchingrequestparameter import MatchingRequestParameter
from odxtools.parameters.nrcconstparameter import NrcConstParameter
from odxtools.parameters.physicalconstantparameter import PhysicalConstantParameter
from odxtools.parameters.reservedparameter import ReservedParameter
from odxtools.parameters.valueparameter import ValueParameter
from odxtools.physicaltype import PhysicalType
from odxtools.request import Request
from odxtools.response import Response, ResponseType
from odxtools.snrefcontext import SnRefContext
from odxtools.standardlengthtype import StandardLengthType
from odxtools.staticfield import StaticField
from odxtools.structure import Structure
from odxtools.termination import Termination
from odxtools.text import Text

doc_frags = [OdxDocFragment("UnitTest", DocType.CONTAINER)]


class TestIdentifyingService(unittest.TestCase):

    def test_prefix_tree_construction(self) -> None:
        diag_coded_type = StandardLengthType(
            base_data_type=DataType.A_UINT32,
            bit_length=8,
        )
        diag_coded_type_2 = StandardLengthType(
            base_data_type=DataType.A_UINT32,
            bit_length=16,
        )
        req_param1 = CodedConstParameter(
            short_name="SID",
            diag_coded_type=diag_coded_type,
            coded_value_raw=str(0x7D),
            byte_position=0,
        )
        req_param2 = CodedConstParameter(
            short_name="coded_const_parameter_2",
            diag_coded_type=diag_coded_type,
            coded_value_raw=str(0xAB),
            byte_position=1,
        )
        req = Request(
            odx_id=OdxLinkId("request_id", doc_frags),
            short_name="request_sn",
            parameters=NamedItemList([req_param1, req_param2]),
        )
        odxlinks = OdxLinkDatabase()
        odxlinks.update({req.odx_id: req})
        service = DiagService(
            odx_id=OdxLinkId("service_id", doc_frags),
            short_name="service_sn",
            request_ref=OdxLinkRef.from_id(req.odx_id),
        )

        req2_param2 = CodedConstParameter(
            short_name="coded_const_parameter_3",
            diag_coded_type=diag_coded_type_2,
            coded_value_raw=str(0xCDE),
        )
        req2 = Request(
            odx_id=OdxLinkId("request_id2", doc_frags),
            short_name="request_sn2",
            parameters=NamedItemList([req_param1, req2_param2]),
        )
        odxlinks.update({req2.odx_id: req2})

        resp2_param2 = CodedConstParameter(
            short_name="coded_const_parameter_4",
            diag_coded_type=diag_coded_type_2,
            coded_value_raw=str(0xC86),
        )
        resp2 = Response(
            odx_id=OdxLinkId("response_id2", doc_frags),
            short_name="response_sn2",
            response_type=ResponseType.NEGATIVE,
            parameters=NamedItemList([req_param1, resp2_param2]),
        )
        odxlinks.update({resp2.odx_id: resp2})

        service2 = DiagService(
            odx_id=OdxLinkId("service_id2", doc_frags),
            short_name="service_sn2",
            request_ref=OdxLinkRef.from_id(req2.odx_id),
            pos_response_refs=[OdxLinkRef.from_id(resp2.odx_id)],
        )

        ecu_variant_raw = EcuVariantRaw(
            variant_type=DiagLayerType.ECU_VARIANT,
            odx_id=OdxLinkId("dl_id", doc_frags),
            short_name="dl_sn",
            diag_comms_raw=[service, service2],
            requests=NamedItemList([req, req2]),
            positive_responses=NamedItemList([resp2]),
        )
        ecu_variant = EcuVariant(diag_layer_raw=ecu_variant_raw)
        db = Database()
        ecu_variant._resolve_odxlinks(odxlinks)
        ecu_variant._finalize_init(db, odxlinks)

        self.assertEqual(
            ecu_variant._prefix_tree,
            {0x7D: {
                0xAB: {
                    -1: [service]
                },
                0xC: {
                    0xDE: {
                        -1: [service2]
                    },
                    0x86: {
                        -1: [service2]
                    }
                }
            }},
        )


class TestDecoding(unittest.TestCase):

    def test_decode_request_coded_const(self) -> None:
        diag_coded_type = StandardLengthType(
            base_data_type=DataType.A_UINT32,
            bit_length=8,
        )
        req_param1 = CodedConstParameter(
            short_name="SID",
            diag_coded_type=diag_coded_type,
            coded_value_raw=str(0x7D),
            byte_position=0,
        )
        req_param2 = CodedConstParameter(
            short_name="coded_const_parameter_2",
            diag_coded_type=diag_coded_type,
            coded_value_raw=str(0xAB),
            byte_position=1,
        )
        req = Request(
            odx_id=OdxLinkId("request_id", doc_frags),
            short_name="request_sn",
            parameters=NamedItemList([req_param1, req_param2]),
        )

        service = DiagService(
            odx_id=OdxLinkId("service_id", doc_frags),
            short_name="service_sn",
            request_ref=OdxLinkRef.from_id(req.odx_id),
        )
        ecu_variant_raw = EcuVariantRaw(
            variant_type=DiagLayerType.ECU_VARIANT,
            odx_id=OdxLinkId("dl_id", doc_frags),
            short_name="dl_sn",
            diag_comms_raw=[service],
            requests=NamedItemList([req]),
        )
        ecu_variant = EcuVariant(diag_layer_raw=ecu_variant_raw)
        odxlinks = OdxLinkDatabase()
        odxlinks.update(ecu_variant._build_odxlinks())
        db = Database()
        ecu_variant._resolve_odxlinks(odxlinks)
        ecu_variant._finalize_init(db, odxlinks)

        coded_message = bytes([0x7D, 0xAB])
        expected_message = Message(
            coded_message=coded_message,
            service=service,
            coding_object=req,
            param_dict={
                "SID": 0x7D,
                "coded_const_parameter_2": 0xAB
            },
        )
        decoded_message = ecu_variant.decode(coded_message)[0]

        self.assertEqual(expected_message.coded_message, decoded_message.coded_message)
        self.assertEqual(expected_message.service, decoded_message.service)
        self.assertEqual(expected_message.coding_object, decoded_message.coding_object)
        self.assertEqual(expected_message.param_dict, decoded_message.param_dict)

    def test_decode_nrc_const(self) -> None:
        diag_coded_type = StandardLengthType(
            base_data_type=DataType.A_UINT32,
            bit_length=8,
        )
        dop = DataObjectProperty(
            odx_id=OdxLinkId("dop.id", doc_frags),
            short_name="dop_sn",
            long_name="example dop",
            diag_coded_type=diag_coded_type,
            physical_type=PhysicalType(base_data_type=DataType.A_UINT32),
            compu_method=IdenticalCompuMethod(
                category=CompuCategory.IDENTICAL,
                internal_type=DataType.A_UINT32,
                physical_type=DataType.A_UINT32),
        )
        param1 = CodedConstParameter(
            short_name="param1",
            diag_coded_type=diag_coded_type,
            coded_value_raw=str(0x12),
            byte_position=0,
        )
        param2 = NrcConstParameter(
            short_name="param2",
            diag_coded_type=diag_coded_type,
            coded_values_raw=[str(0x34), str(0xAB)],
            byte_position=1,
        )
        param3 = ValueParameter(
            short_name="param3",
            dop_ref=OdxLinkRef.from_id(dop.odx_id),
            byte_position=1,
        )
        resp = Response(
            odx_id=OdxLinkId("response_id", doc_frags),
            short_name="response_sn",
            response_type=ResponseType.POSITIVE,
            parameters=NamedItemList([param1, param2, param3]),
        )

        req = Request(
            odx_id=OdxLinkId("request_id", doc_frags),
            short_name="request_sn",
            parameters=NamedItemList([
                CodedConstParameter(
                    short_name="req_param1",
                    diag_coded_type=diag_coded_type,
                    coded_value_raw=str(0xB0),
                    byte_position=0,
                )
            ]),
        )

        service = DiagService(
            odx_id=OdxLinkId("service_id", doc_frags),
            short_name="service_sn",
            request_ref=OdxLinkRef.from_id(req.odx_id),
            neg_response_refs=[OdxLinkRef.from_id(resp.odx_id)],
        )

        base_variant_raw = BaseVariantRaw(
            variant_type=DiagLayerType.BASE_VARIANT,
            odx_id=OdxLinkId("dl_id", doc_frags),
            short_name="dl_sn",
            diag_data_dictionary_spec=DiagDataDictionarySpec(
                data_object_props=NamedItemList([dop])),
            diag_comms_raw=[service],
            requests=NamedItemList([req]),
            negative_responses=NamedItemList([resp]),
        )
        ecu_variant = BaseVariant(diag_layer_raw=base_variant_raw)
        db = Database()
        odxlinks = OdxLinkDatabase()
        odxlinks.update(ecu_variant._build_odxlinks())
        ecu_variant._resolve_odxlinks(odxlinks)
        ecu_variant._finalize_init(db, odxlinks)

        decoded = resp.decode(bytes.fromhex("12ab"))
        self.assertEqual(decoded, {'param1': 0x12, 'param2': 0xab, 'param3': 0xab})

        decoded = resp.decode(bytes.fromhex("1234"))
        self.assertEqual(decoded, {'param1': 0x12, 'param2': 0x34, 'param3': 0x34})

        with self.assertRaises(DecodeMismatch):
            decoded = resp.decode(bytes.fromhex("12bc"))

        decoded_list = ecu_variant.decode_response(bytes.fromhex("12ab"), bytes.fromhex("B0"))
        self.assertEqual(len(decoded_list), 1)
        self.assertEqual(decoded_list[0].param_dict, {
            'param1': 0x12,
            'param2': 0xab,
            'param3': 0xab
        })

        with self.assertRaises(DecodeError):
            decoded_list = ecu_variant.decode_response(bytes.fromhex("12bc"), bytes.fromhex("B0"))

    def test_decode_request_coded_const_undefined_byte_position(self) -> None:
        """Test decoding of parameter
        Test if the decoding works if the byte position of the second parameter
        must be inferred from the order in the surrounding structure."""
        diag_coded_type = StandardLengthType(
            base_data_type=DataType.A_UINT32,
            bit_length=8,
        )
        req_param1 = CodedConstParameter(
            short_name="SID",
            diag_coded_type=diag_coded_type,
            coded_value_raw=str(0x12),
            byte_position=0,
        )
        req_param2 = CodedConstParameter(
            short_name="coded_const_parameter_2",
            diag_coded_type=diag_coded_type,
            coded_value_raw=str(0x78),
            byte_position=3,
        )
        req_param3 = CodedConstParameter(
            short_name="coded_const_parameter_3",
            diag_coded_type=diag_coded_type,
            coded_value_raw=str(0x34),
            byte_position=1,
        )
        req_param4 = CodedConstParameter(
            short_name="coded_const_parameter_4",
            diag_coded_type=diag_coded_type,
            coded_value_raw=str(0x56),
        )
        req = Request(
            odx_id=OdxLinkId("request_id", doc_frags),
            short_name="request_sn",
            parameters=NamedItemList([req_param1, req_param2, req_param3, req_param4]),
        )

        odxlinks = OdxLinkDatabase()
        odxlinks.update({req.odx_id: req})
        service = DiagService(
            odx_id=OdxLinkId("service_id", doc_frags),
            short_name="service_sn",
            request_ref=OdxLinkRef.from_id(req.odx_id),
        )
        ecu_variant_raw = EcuVariantRaw(
            variant_type=DiagLayerType.ECU_VARIANT,
            odx_id=OdxLinkId("dl_id", doc_frags),
            short_name="dl_sn",
            diag_comms_raw=[service],
            requests=NamedItemList([req]),
        )
        ecu_variant = EcuVariant(diag_layer_raw=ecu_variant_raw)
        db = Database()
        ecu_variant._resolve_odxlinks(odxlinks)
        ecu_variant._finalize_init(db, odxlinks)

        self.assertDictEqual(ecu_variant._prefix_tree,
                             {0x12: {
                                 0x34: {
                                     0x56: {
                                         0x78: {
                                             -1: [service]
                                         }
                                     }
                                 }
                             }})

        coded_message = bytes([0x12, 0x34, 0x56, 0x78])
        expected_message = Message(
            coded_message=coded_message,
            service=service,
            coding_object=req,
            param_dict={
                "SID": 0x12,
                "coded_const_parameter_2": 0x78,
                "coded_const_parameter_3": 0x34,
                "coded_const_parameter_4": 0x56,
            },
        )
        decoded_message = ecu_variant.decode(coded_message)[0]
        self.assertEqual(expected_message.coded_message, decoded_message.coded_message)
        self.assertEqual(expected_message.service, decoded_message.service)
        self.assertEqual(expected_message.coding_object, decoded_message.coding_object)
        self.assertEqual(expected_message.param_dict, decoded_message.param_dict)

    def test_decode_request_structure(self) -> None:
        """Test the decoding for a structure."""
        diag_coded_type = StandardLengthType(
            base_data_type=DataType.A_UINT32,
            bit_length=8,
        )
        diag_coded_type_4 = StandardLengthType(
            base_data_type=DataType.A_UINT32,
            bit_length=4,
        )

        compu_method = IdenticalCompuMethod(
            category=CompuCategory.IDENTICAL,
            internal_type=DataType.A_INT32,
            physical_type=DataType.A_INT32)
        dop = DataObjectProperty(
            odx_id=OdxLinkId("dop.odx_id", doc_frags),
            short_name="dop_sn",
            diag_coded_type=diag_coded_type_4,
            physical_type=PhysicalType(base_data_type=DataType.A_UINT32),
            compu_method=compu_method,
        )

        req_param1 = CodedConstParameter(
            short_name="SID",
            diag_coded_type=diag_coded_type,
            coded_value_raw=str(0x12),
            byte_position=0,
        )

        struct_param1 = CodedConstParameter(
            short_name="struct_param_1",
            diag_coded_type=diag_coded_type_4,
            coded_value_raw=str(0x4),
            byte_position=0,
            bit_position=0,
        )
        struct_param2 = ValueParameter(
            short_name="struct_param_2",
            dop_ref=OdxLinkRef.from_id(dop.odx_id),
            byte_position=0,
            bit_position=4,
        )
        struct = Structure(
            odx_id=OdxLinkId("struct_id", doc_frags),
            short_name="struct",
            parameters=NamedItemList([struct_param1, struct_param2]),
        )
        req_param2 = ValueParameter(
            short_name="structured_param",
            dop_ref=OdxLinkRef.from_id(struct.odx_id),
        )

        req = Request(
            odx_id=OdxLinkId("request_id", doc_frags),
            short_name="request_sn",
            parameters=NamedItemList([req_param1, req_param2]),
        )
        service = DiagService(
            odx_id=OdxLinkId("service_id", doc_frags),
            short_name="service_sn",
            request_ref=OdxLinkRef.from_id(req.odx_id),
        )
        ecu_variant_raw = EcuVariantRaw(
            variant_type=DiagLayerType.ECU_VARIANT,
            odx_id=OdxLinkId("dl_id", doc_frags),
            short_name="dl_sn",
            diag_data_dictionary_spec=DiagDataDictionarySpec(
                data_object_props=NamedItemList([dop]), structures=NamedItemList([struct])),
            diag_comms_raw=[service],
            requests=NamedItemList([req]),
        )
        ecu_variant = EcuVariant(diag_layer_raw=ecu_variant_raw)
        odxlinks = OdxLinkDatabase()
        odxlinks.update(ecu_variant._build_odxlinks())
        db = Database()
        ecu_variant._resolve_odxlinks(odxlinks)
        ecu_variant._finalize_init(db, odxlinks)

        coded_message = bytes([0x12, 0x34])
        expected_message = Message(
            coded_message=coded_message,
            service=service,
            coding_object=req,
            param_dict={
                "SID": 0x12,
                "structured_param": {
                    "struct_param_1": 4,
                    "struct_param_2": 3
                },
            },
        )
        decoded_message = ecu_variant.decode(coded_message)[0]
        self.assertEqual(expected_message.coded_message, decoded_message.coded_message)
        self.assertEqual(expected_message.service, decoded_message.service)
        self.assertEqual(expected_message.coding_object, decoded_message.coding_object)
        self.assertEqual(expected_message.param_dict, decoded_message.param_dict)

    def test_static_field_coding(self) -> None:
        """Test en- and decoding of static fields."""
        diag_coded_type = StandardLengthType(
            base_data_type=DataType.A_UINT32,
            bit_length=8,
        )

        compu_method = IdenticalCompuMethod(
            category=CompuCategory.IDENTICAL,
            internal_type=DataType.A_INT32,
            physical_type=DataType.A_INT32)
        dop = DataObjectProperty(
            odx_id=OdxLinkId("static_field.dop.id", doc_frags),
            short_name="static_field_dop_sn",
            diag_coded_type=diag_coded_type,
            physical_type=PhysicalType(base_data_type=DataType.A_UINT32),
            compu_method=compu_method,
        )

        req_param1 = CodedConstParameter(
            short_name="SID",
            diag_coded_type=diag_coded_type,
            coded_value_raw=str(0x12),
            byte_position=0,
        )

        struct_param1 = ValueParameter(
            short_name="static_field_struct_param_1",
            dop_ref=OdxLinkRef.from_id(dop.odx_id),
        )
        struct_param2 = ValueParameter(
            short_name="static_field_struct_param_2",
            dop_ref=OdxLinkRef.from_id(dop.odx_id),
        )
        struct = Structure(
            odx_id=OdxLinkId("static_field_struct.id", doc_frags),
            short_name="static_field_struct",
            parameters=NamedItemList([struct_param1, struct_param2]),
        )
        static_field = StaticField(
            odx_id=OdxLinkId("static_field.id", doc_frags),
            short_name="static_field_sn",
            structure_ref=OdxLinkRef.from_id(struct.odx_id),
            is_visible_raw=True,
            fixed_number_of_items=2,
            item_byte_size=3,
        )
        req_param2 = ValueParameter(
            short_name="static_field_param",
            dop_ref=OdxLinkRef.from_id(static_field.odx_id),
        )

        req = Request(
            odx_id=OdxLinkId("static_field.request.id", doc_frags),
            short_name="static_field_request_sn",
            parameters=NamedItemList([req_param1, req_param2]),
        )
        service = DiagService(
            odx_id=OdxLinkId("static_field.service.id", doc_frags),
            short_name="static_field_service_sn",
            request_ref=OdxLinkRef.from_id(req.odx_id),
        )
        ecu_variant_raw = EcuVariantRaw(
            variant_type=DiagLayerType.ECU_VARIANT,
            odx_id=OdxLinkId("dl.id", doc_frags),
            short_name="dl_sn",
            diag_data_dictionary_spec=DiagDataDictionarySpec(
                data_object_props=NamedItemList([dop]),
                structures=NamedItemList([struct]),
                static_fields=NamedItemList([static_field])),
            diag_comms_raw=[service],
            requests=NamedItemList([req]),
        )
        ecu_variant = EcuVariant(diag_layer_raw=ecu_variant_raw)
        odxlinks = OdxLinkDatabase()
        odxlinks.update(ecu_variant._build_odxlinks())
        db = Database()
        ecu_variant._resolve_odxlinks(odxlinks)
        ecu_variant._finalize_init(db, odxlinks)

        expected_message = Message(
            coded_message=bytes([0x12, 0x34, 0x56, 0x00, 0x78, 0x9a, 0x00]),
            service=service,
            coding_object=req,
            param_dict={
                "SID":
                    0x12,
                "static_field_param": [
                    {
                        "static_field_struct_param_1": 0x34,
                        "static_field_struct_param_2": 0x56
                    },
                    {
                        "static_field_struct_param_1": 0x78,
                        "static_field_struct_param_2": 0x9a
                    },
                ],
            },
        )

        # test encoding
        encoded_message = ecu_variant.services.static_field_service_sn(
            **expected_message.param_dict)
        self.assertEqual(encoded_message, expected_message.coded_message)

        # test decoding
        decoded_message = ecu_variant.decode(expected_message.coded_message)[0]
        self.assertEqual(expected_message.coded_message, decoded_message.coded_message)
        self.assertEqual(expected_message.service, decoded_message.service)
        self.assertEqual(expected_message.coding_object, decoded_message.coding_object)
        self.assertEqual(expected_message.param_dict, decoded_message.param_dict)

    def test_dynamic_endmarker_field_coding(self) -> None:
        """Test en- and decoding of dynamic endmarker fields."""
        diag_coded_type = StandardLengthType(
            base_data_type=DataType.A_UINT32,
            bit_length=8,
        )
        diag_coded_type_4 = StandardLengthType(
            base_data_type=DataType.A_UINT32,
            bit_length=4,
        )
        diag_coded_endmarker_type = StandardLengthType(
            base_data_type=DataType.A_BYTEFIELD,
            bit_length=24,
        )

        compu_method = IdenticalCompuMethod(
            category=CompuCategory.IDENTICAL,
            internal_type=DataType.A_INT32,
            physical_type=DataType.A_INT32)
        compu_method_bytefield = IdenticalCompuMethod(
            category=CompuCategory.IDENTICAL,
            internal_type=DataType.A_BYTEFIELD,
            physical_type=DataType.A_BYTEFIELD)

        dop = DataObjectProperty(
            odx_id=OdxLinkId("demf.dop.id", doc_frags),
            short_name="demf_dop_sn",
            diag_coded_type=diag_coded_type_4,
            physical_type=PhysicalType(base_data_type=DataType.A_UINT32),
            compu_method=compu_method,
        )
        dyn_end_dop = DataObjectProperty(
            odx_id=OdxLinkId("demf.end_dop.id", doc_frags),
            short_name="demf_end_dop_sn",
            diag_coded_type=diag_coded_endmarker_type,
            physical_type=PhysicalType(base_data_type=DataType.A_BYTEFIELD),
            compu_method=compu_method_bytefield,
        )
        dyn_end_dop_ref = DynEndDopRef(
            termination_value_raw="ffffff",
            ref_id=dyn_end_dop.odx_id.local_id,
            ref_docs=dyn_end_dop.odx_id.doc_fragments,
        )

        req_param1 = CodedConstParameter(
            short_name="SID",
            diag_coded_type=diag_coded_type,
            coded_value_raw=str(0x12),
            byte_position=0,
        )
        req_param1_1 = CodedConstParameter(
            short_name="SID",
            diag_coded_type=diag_coded_type,
            coded_value_raw=str(0x13),
            byte_position=0,
        )

        struct_param1 = CodedConstParameter(
            short_name="struct_param_1",
            diag_coded_type=diag_coded_type_4,
            coded_value_raw=str(0x4),
            byte_position=0,
            bit_position=0,
        )
        struct_param2 = ValueParameter(
            short_name="struct_param_2",
            dop_ref=OdxLinkRef.from_id(dop.odx_id),
            byte_position=0,
            bit_position=4,
        )
        struct = Structure(
            odx_id=OdxLinkId("demf_struct.id", doc_frags),
            short_name="demf_struct",
            parameters=NamedItemList([struct_param1, struct_param2]),
        )
        demf = DynamicEndmarkerField(
            odx_id=OdxLinkId("demf.id", doc_frags),
            short_name="demf_sn",
            structure_ref=OdxLinkRef.from_id(struct.odx_id),
            is_visible_raw=True,
            dyn_end_dop_ref=dyn_end_dop_ref,
        )
        req_param2 = ValueParameter(
            short_name="demf_param",
            dop_ref=OdxLinkRef.from_id(demf.odx_id),
        )
        req_demf_endmarker_param = ReservedParameter(
            short_name="demf_endmarker",
            bit_length=24,
        )
        req_param3 = CodedConstParameter(
            short_name="demf_post_param",
            diag_coded_type=diag_coded_type,
            coded_value_raw=str(0xcc),
        )

        req = Request(
            odx_id=OdxLinkId("demf.request.id", doc_frags),
            short_name="demf_request_sn",
            parameters=NamedItemList([req_param1, req_param2, req_demf_endmarker_param,
                                      req_param3]),
        )

        # same as the previous request, but the dynamic endmarker
        # field is at the end of the PDU, so no endmarker is added
        req_end_of_pdu = Request(
            odx_id=OdxLinkId("demf_eopdu.request.id", doc_frags),
            short_name="demf_eopdu_request_sn",
            parameters=NamedItemList([req_param1_1, req_param2]),
        )

        service = DiagService(
            odx_id=OdxLinkId("demf.service.id", doc_frags),
            short_name="demf_service_sn",
            request_ref=OdxLinkRef.from_id(req.odx_id),
        )
        service_eopdu = DiagService(
            odx_id=OdxLinkId("demf.service_eopdu.id", doc_frags),
            short_name="demf_service_eopdu_sn",
            request_ref=OdxLinkRef.from_id(req_end_of_pdu.odx_id),
        )
        ecu_variant_raw = EcuVariantRaw(
            variant_type=DiagLayerType.ECU_VARIANT,
            odx_id=OdxLinkId("dl.id", doc_frags),
            short_name="dl_sn",
            diag_data_dictionary_spec=DiagDataDictionarySpec(
                data_object_props=NamedItemList([dop, dyn_end_dop]),
                structures=NamedItemList([struct]),
                dynamic_endmarker_fields=NamedItemList([demf])),
            diag_comms_raw=[service, service_eopdu],
            requests=NamedItemList([req, req_end_of_pdu]),
        )
        ecu_variant = EcuVariant(diag_layer_raw=ecu_variant_raw)
        odxlinks = OdxLinkDatabase()
        odxlinks.update(ecu_variant._build_odxlinks())
        db = Database()
        ecu_variant._resolve_odxlinks(odxlinks)
        ecu_variant._finalize_init(db, odxlinks)

        ######
        ## test with endmarker termination
        expected_message = Message(
            coded_message=bytes([0x12, 0x34, 0x44, 0x54, 0xff, 0xff, 0xff, 0xcc]),
            service=service,
            coding_object=req,
            param_dict={
                "SID": 0x12,
                "demf_param": [
                    {
                        "struct_param_1": 4,
                        "struct_param_2": 3
                    },
                    {
                        "struct_param_1": 4,
                        "struct_param_2": 4
                    },
                    {
                        "struct_param_1": 4,
                        "struct_param_2": 5
                    },
                ],
                "demf_endmarker": 0xffffff,
                "demf_post_param": 0xcc,
            },
        )

        # test encoding
        encoded_message = ecu_variant.services.demf_service_sn(**expected_message.param_dict)
        self.assertEqual(encoded_message, expected_message.coded_message)

        # test decoding
        decoded_message = ecu_variant.decode(expected_message.coded_message)[0]
        self.assertEqual(expected_message.coded_message, decoded_message.coded_message)
        self.assertEqual(expected_message.service, decoded_message.service)
        self.assertEqual(expected_message.coding_object, decoded_message.coding_object)
        self.assertEqual(expected_message.param_dict, decoded_message.param_dict)

        ######
        ## test with end-of-pdu termination
        expected_message_eopdu = Message(
            coded_message=bytes([0x13, 0x34, 0x44, 0x54]),
            service=service_eopdu,
            coding_object=req_end_of_pdu,
            param_dict={
                "SID":
                    0x13,
                "demf_param": [
                    {
                        "struct_param_1": 4,
                        "struct_param_2": 3
                    },
                    {
                        "struct_param_1": 4,
                        "struct_param_2": 4
                    },
                    {
                        "struct_param_1": 4,
                        "struct_param_2": 5
                    },
                ],
            },
        )

        # test encoding
        encoded_message = ecu_variant.services.demf_service_eopdu_sn(
            **expected_message_eopdu.param_dict)
        self.assertEqual(encoded_message, expected_message_eopdu.coded_message)

        # test decoding
        decoded_message = ecu_variant.decode(expected_message_eopdu.coded_message)[0]
        self.assertEqual(expected_message_eopdu.coded_message, decoded_message.coded_message)
        self.assertEqual(expected_message_eopdu.service, decoded_message.service)
        self.assertEqual(expected_message_eopdu.coding_object, decoded_message.coding_object)
        self.assertEqual(expected_message_eopdu.param_dict, decoded_message.param_dict)

    def test_dynamic_length_field_coding(self) -> None:
        """Test en- and decoding of dynamic length fields."""
        diag_coded_type = StandardLengthType(
            base_data_type=DataType.A_UINT32,
            bit_length=8,
        )
        diag_coded_type_4 = StandardLengthType(
            base_data_type=DataType.A_UINT32,
            bit_length=4,
        )

        compu_method = IdenticalCompuMethod(
            category=CompuCategory.IDENTICAL,
            internal_type=DataType.A_INT32,
            physical_type=DataType.A_INT32)
        dop = DataObjectProperty(
            odx_id=OdxLinkId("dlf.dop.id", doc_frags),
            short_name="dlf_dop_sn",
            diag_coded_type=diag_coded_type_4,
            physical_type=PhysicalType(base_data_type=DataType.A_UINT32),
            compu_method=compu_method,
        )

        req_param1 = CodedConstParameter(
            short_name="SID",
            diag_coded_type=diag_coded_type,
            coded_value_raw=str(0x12),
            byte_position=0,
        )

        struct_param1 = CodedConstParameter(
            short_name="struct_param_1",
            diag_coded_type=diag_coded_type_4,
            coded_value_raw=str(0x4),
            byte_position=0,
            bit_position=0,
        )
        struct_param2 = ValueParameter(
            short_name="struct_param_2",
            dop_ref=OdxLinkRef.from_id(dop.odx_id),
            byte_position=0,
            bit_position=4,
        )
        struct = Structure(
            odx_id=OdxLinkId("dlf_struct.id", doc_frags),
            short_name="dlf_struct",
            parameters=NamedItemList([struct_param1, struct_param2]),
        )
        det_num_items = DetermineNumberOfItems(
            byte_position=1, bit_position=3, dop_ref=OdxLinkRef.from_id(dop.odx_id))
        dlf = DynamicLengthField(
            odx_id=OdxLinkId("dlf.id", doc_frags),
            short_name="dlf_sn",
            structure_ref=OdxLinkRef.from_id(struct.odx_id),
            is_visible_raw=True,
            offset=3,
            determine_number_of_items=det_num_items,
        )
        req_param2 = ValueParameter(
            short_name="dlf_param",
            dop_ref=OdxLinkRef.from_id(dlf.odx_id),
        )

        req = Request(
            odx_id=OdxLinkId("dlf.request.id", doc_frags),
            short_name="dlf_request_sn",
            parameters=NamedItemList([req_param1, req_param2]),
        )
        service = DiagService(
            odx_id=OdxLinkId("dlf.service.id", doc_frags),
            short_name="dlf_service_sn",
            request_ref=OdxLinkRef.from_id(req.odx_id),
        )
        ecu_variant_raw = EcuVariantRaw(
            variant_type=DiagLayerType.ECU_VARIANT,
            odx_id=OdxLinkId("dl.id", doc_frags),
            short_name="dl_sn",
            diag_data_dictionary_spec=DiagDataDictionarySpec(
                data_object_props=NamedItemList([dop]),
                structures=NamedItemList([struct]),
                dynamic_length_fields=NamedItemList([dlf])),
            diag_comms_raw=[service],
            requests=NamedItemList([req]),
        )
        ecu_variant = EcuVariant(diag_layer_raw=ecu_variant_raw)
        odxlinks = OdxLinkDatabase()
        odxlinks.update(ecu_variant._build_odxlinks())
        db = Database()
        ecu_variant._resolve_odxlinks(odxlinks)
        ecu_variant._finalize_init(db, odxlinks)

        expected_message = Message(
            coded_message=bytes([0x12, 0x00, 0x18, 0x00, 0x34, 0x44, 0x54]),
            service=service,
            coding_object=req,
            param_dict={
                "SID":
                    0x12,
                "dlf_param": [
                    {
                        "struct_param_1": 4,
                        "struct_param_2": 3
                    },
                    {
                        "struct_param_1": 4,
                        "struct_param_2": 4
                    },
                    {
                        "struct_param_1": 4,
                        "struct_param_2": 5
                    },
                ],
            },
        )

        # test encoding
        encoded_message = ecu_variant.services.dlf_service_sn(**expected_message.param_dict)
        self.assertEqual(encoded_message, expected_message.coded_message)

        # test decoding
        decoded_message = ecu_variant.decode(expected_message.coded_message)[0]
        self.assertEqual(expected_message.coded_message, decoded_message.coded_message)
        self.assertEqual(expected_message.service, decoded_message.service)
        self.assertEqual(expected_message.coding_object, decoded_message.coding_object)
        self.assertEqual(expected_message.param_dict, decoded_message.param_dict)

    def test_decode_request_end_of_pdu_field(self) -> None:
        """Test decoding of end-of-pdu fields."""
        diag_coded_type = StandardLengthType(
            base_data_type=DataType.A_UINT32,
            bit_length=8,
        )
        diag_coded_type_4 = StandardLengthType(
            base_data_type=DataType.A_UINT32,
            bit_length=4,
        )

        compu_method = IdenticalCompuMethod(
            category=CompuCategory.IDENTICAL,
            internal_type=DataType.A_INT32,
            physical_type=DataType.A_INT32)
        dop = DataObjectProperty(
            odx_id=OdxLinkId("dop.id", doc_frags),
            short_name="dop_sn",
            diag_coded_type=diag_coded_type_4,
            physical_type=PhysicalType(base_data_type=DataType.A_UINT32),
            compu_method=compu_method,
        )

        req_param1 = CodedConstParameter(
            short_name="SID",
            diag_coded_type=diag_coded_type,
            coded_value_raw=str(0x12),
            byte_position=0,
        )

        struct_param1 = CodedConstParameter(
            short_name="struct_param_1",
            diag_coded_type=diag_coded_type_4,
            coded_value_raw=str(0x4),
            byte_position=0,
            bit_position=0,
        )
        struct_param2 = ValueParameter(
            short_name="struct_param_2",
            dop_ref=OdxLinkRef.from_id(dop.odx_id),
            byte_position=0,
            bit_position=4,
        )
        struct = Structure(
            odx_id=OdxLinkId("struct_id", doc_frags),
            short_name="struct",
            parameters=NamedItemList([struct_param1, struct_param2]),
        )
        eopf = EndOfPduField(
            odx_id=OdxLinkId("eopf_id", doc_frags),
            short_name="eopf_sn",
            structure_ref=OdxLinkRef.from_id(struct.odx_id),
            is_visible_raw=True,
        )
        req_param2 = ValueParameter(
            short_name="eopf_param",
            dop_ref=OdxLinkRef.from_id(eopf.odx_id),
        )

        req = Request(
            odx_id=OdxLinkId("request_id", doc_frags),
            short_name="request_sn",
            parameters=NamedItemList([req_param1, req_param2]),
        )
        service = DiagService(
            odx_id=OdxLinkId("service_id", doc_frags),
            short_name="service_sn",
            request_ref=OdxLinkRef.from_id(req.odx_id),
        )
        ecu_variant_raw = EcuVariantRaw(
            variant_type=DiagLayerType.ECU_VARIANT,
            odx_id=OdxLinkId("dl_id", doc_frags),
            short_name="dl_sn",
            diag_data_dictionary_spec=DiagDataDictionarySpec(
                data_object_props=NamedItemList([dop]),
                structures=NamedItemList([struct]),
                end_of_pdu_fields=NamedItemList([eopf])),
            diag_comms_raw=[service],
            requests=NamedItemList([req]),
        )
        ecu_variant = EcuVariant(diag_layer_raw=ecu_variant_raw)
        odxlinks = OdxLinkDatabase()
        odxlinks.update(ecu_variant._build_odxlinks())
        db = Database()
        ecu_variant._resolve_odxlinks(odxlinks)
        ecu_variant._finalize_init(db, odxlinks)

        coded_message = bytes([0x12, 0x34, 0x54])
        expected_message = Message(
            coded_message=coded_message,
            service=service,
            coding_object=req,
            param_dict={
                "SID":
                    0x12,
                "eopf_param": [
                    {
                        "struct_param_1": 4,
                        "struct_param_2": 3
                    },
                    {
                        "struct_param_1": 4,
                        "struct_param_2": 5
                    },
                ],
            },
        )
        decoded_message = ecu_variant.decode(coded_message)[0]
        self.assertEqual(expected_message.coded_message, decoded_message.coded_message)
        self.assertEqual(expected_message.service, decoded_message.service)
        self.assertEqual(expected_message.coding_object, decoded_message.coding_object)
        self.assertEqual(expected_message.param_dict, decoded_message.param_dict)

    def test_decode_request_linear_compu_method(self) -> None:
        compu_method = LinearCompuMethod(
            category=CompuCategory.LINEAR,
            compu_internal_to_phys=CompuInternalToPhys(compu_scales=[
                CompuScale(
                    compu_rational_coeffs=CompuRationalCoeffs(
                        value_type=DataType.A_INT32,
                        numerators=[1, 5],
                        denominators=[1],
                    ),
                    domain_type=DataType.A_INT32,
                    range_type=DataType.A_INT32),
            ]),
            internal_type=DataType.A_INT32,
            physical_type=DataType.A_INT32,
        )
        diag_coded_type = StandardLengthType(
            base_data_type=DataType.A_UINT32,
            bit_length=8,
        )
        dop = DataObjectProperty(
            odx_id=OdxLinkId("linear.dop.id", doc_frags),
            short_name="linear_dop_sn",
            diag_coded_type=diag_coded_type,
            physical_type=PhysicalType(base_data_type=DataType.A_UINT32),
            compu_method=compu_method,
        )
        req_param1 = CodedConstParameter(
            short_name="SID",
            diag_coded_type=diag_coded_type,
            coded_value_raw=str(0x7D),
            byte_position=0,
        )
        req_param2 = ValueParameter(
            short_name="value_parameter_2",
            dop_ref=OdxLinkRef.from_id(dop.odx_id),
            byte_position=1,
        )
        req = Request(
            odx_id=OdxLinkId("request_id", doc_frags),
            short_name="request_sn",
            parameters=NamedItemList([req_param1, req_param2]),
        )

        service = DiagService(
            odx_id=OdxLinkId("service_id", doc_frags),
            short_name="service_sn",
            request_ref=OdxLinkRef.from_id(req.odx_id),
        )
        ecu_variant_raw = EcuVariantRaw(
            variant_type=DiagLayerType.ECU_VARIANT,
            odx_id=OdxLinkId("dl_id", doc_frags),
            short_name="dl_sn",
            diag_data_dictionary_spec=DiagDataDictionarySpec(
                data_object_props=NamedItemList([dop])),
            diag_comms_raw=[service],
            requests=NamedItemList([req]),
        )
        ecu_variant = EcuVariant(diag_layer_raw=ecu_variant_raw)
        odxlinks = OdxLinkDatabase()
        odxlinks.update(ecu_variant._build_odxlinks())
        db = Database()
        ecu_variant._resolve_odxlinks(odxlinks)
        ecu_variant._finalize_init(db, odxlinks)

        coded_message = bytes([0x7D, 0x12])
        # The physical value of the second parameter is decode(0x12) = decode(18) = 5 * 18 + 1 = 91
        expected_message = Message(
            coded_message=coded_message,
            service=service,
            coding_object=req,
            param_dict={
                "SID": 0x7D,
                "value_parameter_2": 91
            },
        )
        decoded_message = ecu_variant.decode(coded_message)[0]
        self.assertEqual(expected_message.coded_message, decoded_message.coded_message)
        self.assertEqual(expected_message.service, decoded_message.service)
        self.assertEqual(expected_message.coding_object, decoded_message.coding_object)
        self.assertEqual(expected_message.param_dict, decoded_message.param_dict)

    def test_decode_response(self) -> None:
        diag_coded_type = StandardLengthType(
            base_data_type=DataType.A_UINT32,
            bit_length=8,
        )
        req_param1 = CodedConstParameter(
            short_name="SID",
            diag_coded_type=diag_coded_type,
            coded_value_raw=str(0x12),
            byte_position=0,
        )
        req_param2 = CodedConstParameter(
            short_name="req_param",
            diag_coded_type=diag_coded_type,
            coded_value_raw=str(0xAB),
            byte_position=1,
        )
        req = Request(
            odx_id=OdxLinkId("request_id", doc_frags),
            short_name="request_sn",
            parameters=NamedItemList([req_param1, req_param2]),
        )

        resp_param1 = CodedConstParameter(
            short_name="SID",
            diag_coded_type=diag_coded_type,
            coded_value_raw=str(0x34),
            byte_position=0,
        )
        resp_param2 = MatchingRequestParameter(
            short_name="matching_req_param",
            request_byte_position=1,
            byte_length=1,
        )
        pos_response = Response(
            odx_id=OdxLinkId("pos_response_id", doc_frags),
            short_name="pos_response_sn",
            response_type=ResponseType.POSITIVE,
            parameters=NamedItemList([resp_param1, resp_param2]),
        )

        resp_param1 = CodedConstParameter(
            short_name="SID",
            diag_coded_type=diag_coded_type,
            coded_value_raw=str(0x56),
            byte_position=0,
        )
        resp_param2 = MatchingRequestParameter(
            short_name="matching_req_param",
            request_byte_position=1,
            byte_length=1,
        )
        neg_response = Response(
            odx_id=OdxLinkId("neg_response_id", doc_frags),
            short_name="neg_response_sn",
            response_type=ResponseType.NEGATIVE,
            parameters=NamedItemList([resp_param1, resp_param2]),
        )

        service = DiagService(
            odx_id=OdxLinkId("service_id", doc_frags),
            short_name="service_sn",
            request_ref=OdxLinkRef.from_id(req.odx_id),
            pos_response_refs=[OdxLinkRef.from_id(pos_response.odx_id)],
            neg_response_refs=[OdxLinkRef.from_id(neg_response.odx_id)],
        )
        ecu_variant_raw = EcuVariantRaw(
            variant_type=DiagLayerType.ECU_VARIANT,
            odx_id=OdxLinkId("dl_id", doc_frags),
            short_name="dl_sn",
            diag_comms_raw=[service],
            requests=NamedItemList([req]),
            positive_responses=NamedItemList([pos_response]),
            negative_responses=NamedItemList([neg_response]),
        )
        ecu_variant = EcuVariant(diag_layer_raw=ecu_variant_raw)
        odxlinks = OdxLinkDatabase()
        odxlinks.update(ecu_variant._build_odxlinks())
        db = Database()
        ecu_variant._resolve_odxlinks(odxlinks)
        ecu_variant._finalize_init(db, odxlinks)

        for sid, message in [(0x34, pos_response), (0x56, neg_response)]:
            coded_message = bytes([sid, 0xAB])
            expected_message = Message(
                coded_message=coded_message,
                service=service,
                coding_object=message,
                param_dict={
                    "SID": sid,
                    "matching_req_param": 0xAB
                },
            )
            decoded_message = ecu_variant.decode(coded_message)[0]
            self.assertEqual(expected_message.coded_message, decoded_message.coded_message)
            self.assertEqual(expected_message.service, decoded_message.service)
            self.assertEqual(expected_message.coding_object, decoded_message.coding_object)
            self.assertEqual(expected_message.param_dict, decoded_message.param_dict)

        # check object serialization and deserialization using pickle
        pickled_ecu_var = pickle.dumps(ecu_variant)
        unpickled_ecu_var = pickle.loads(pickled_ecu_var)

        self.assertEqual(ecu_variant, unpickled_ecu_var)

    def test_code_dtc(self) -> None:
        odxlinks = OdxLinkDatabase()
        diag_coded_type = StandardLengthType(
            base_data_type=DataType.A_UINT32,
            bit_length=8,
        )
        compu_method = IdenticalCompuMethod(
            category=CompuCategory.IDENTICAL,
            internal_type=DataType.A_INT32,
            physical_type=DataType.A_INT32)

        dtc1 = DiagnosticTroubleCode(
            odx_id=OdxLinkId("dtcID1", doc_frags),
            short_name="P34_sn",
            trouble_code=0x34,
            text=Text.from_string("Error encountered"),
            display_trouble_code="P34",
        )

        dtc2 = DiagnosticTroubleCode(
            odx_id=OdxLinkId("dtcID2", doc_frags),
            short_name="P56_sn",
            trouble_code=0x56,
            text=Text.from_string("Crashed into wall"),
            display_trouble_code="P56",
        )
        dop = DtcDop(
            odx_id=OdxLinkId("dtc.dop.odx_id", doc_frags),
            short_name="dtc_dop_sn",
            diag_coded_type=diag_coded_type,
            physical_type=PhysicalType(base_data_type=DataType.A_UINT32),
            compu_method=compu_method,
            dtcs_raw=[dtc1, dtc2],
        )
        odxlinks.update(dop._build_odxlinks())
        resp_param1 = CodedConstParameter(
            short_name="SID",
            diag_coded_type=diag_coded_type,
            coded_value_raw=str(0x12),
            byte_position=0,
        )
        resp_param2 = ValueParameter(
            short_name="DTC_Param",
            dop_ref=OdxLinkRef.from_id(dop.odx_id),
            byte_position=1,
        )
        pos_response = Response(
            odx_id=OdxLinkId("pos_response_id", doc_frags),
            short_name="pos_response_sn",
            parameters=NamedItemList([resp_param1, resp_param2]),
            response_type=ResponseType.POSITIVE,
        )
        odxlinks.update(pos_response._build_odxlinks())

        dop._resolve_odxlinks(odxlinks)
        pos_response._resolve_odxlinks(odxlinks)

        expected_coded_message = bytes([0x12, 0x34])
        expected_param_dict: ParameterValueDict = {"SID": 0x12, "DTC_Param": dtc1}

        actual_param_dict = pos_response.decode(expected_coded_message)
        self.assertEqual(actual_param_dict, expected_param_dict)

        actual_coded_message = pos_response.encode(coded_request=None, **expected_param_dict)
        self.assertEqual(actual_coded_message, expected_coded_message)


class TestDecodingAndEncoding(unittest.TestCase):

    def setUp(self) -> None:
        super().setUp()
        odxlinks = OdxLinkDatabase()
        self.dop_bytes_termination_end_of_pdu = DataObjectProperty(
            odx_id=OdxLinkId("DOP_ID", doc_frags),
            short_name="DOP",
            diag_coded_type=MinMaxLengthType(
                base_data_type=DataType.A_BYTEFIELD,
                min_length=0,
                termination=Termination.END_OF_PDU,
            ),
            physical_type=PhysicalType(base_data_type=DataType.A_BYTEFIELD),
            compu_method=IdenticalCompuMethod(
                category=CompuCategory.IDENTICAL,
                internal_type=DataType.A_BYTEFIELD,
                physical_type=DataType.A_BYTEFIELD),
        )
        dop = self.dop_bytes_termination_end_of_pdu
        odxlinks.update(dop._build_odxlinks())
        self.parameter_termination_end_of_pdu = ValueParameter(
            short_name="min_max_parameter",
            dop_ref=OdxLinkRef.from_id(dop.odx_id),
        )

        self.parameter_sid = CodedConstParameter(
            short_name="SID",
            diag_coded_type=StandardLengthType(
                base_data_type=DataType.A_UINT32,
                bit_length=8,
            ),
            coded_value_raw=str(0x12),
            byte_position=0,
        )

        self.parameter_termination_end_of_pdu._resolve_odxlinks(odxlinks)
        self.parameter_sid._resolve_odxlinks(odxlinks)

        snref_ctx = SnRefContext()
        self.parameter_termination_end_of_pdu._resolve_snrefs(snref_ctx)
        self.parameter_sid._resolve_snrefs(snref_ctx)

    def test_min_max_length_type_end_of_pdu(self) -> None:
        req_param1 = self.parameter_sid
        req_param2 = self.parameter_termination_end_of_pdu
        request = Request(
            odx_id=OdxLinkId("request", doc_frags),
            short_name="Request",
            parameters=NamedItemList([req_param1, req_param2]),
        )
        expected_coded_message = bytes([0x12, 0x34])
        expected_param_dict = {"SID": 0x12, "min_max_parameter": bytes([0x34])}

        actual_param_dict = request.decode(expected_coded_message)
        self.assertEqual(dict(actual_param_dict), expected_param_dict)

        actual_coded_message = request.encode(**expected_param_dict)  # type: ignore[arg-type]
        self.assertEqual(actual_coded_message, expected_coded_message)

    def test_min_max_length_type_end_of_pdu_in_structure(self) -> None:
        odxlinks = OdxLinkDatabase()

        struct_param = self.parameter_termination_end_of_pdu

        structure = Structure(
            odx_id=OdxLinkId("structure_id", doc_frags),
            short_name="Structure_with_End_of_PDU_termination",
            parameters=NamedItemList([struct_param]),
        )
        odxlinks.update(structure._build_odxlinks())

        req_param1 = self.parameter_sid
        req_param2 = ValueParameter(
            short_name="min_max_parameter",
            dop_ref=OdxLinkRef.from_id(structure.odx_id),
        )

        request = Request(
            odx_id=OdxLinkId("request", doc_frags),
            short_name="Request",
            parameters=NamedItemList([req_param1, req_param2]),
        )

        req_param1._resolve_odxlinks(odxlinks)
        req_param2._resolve_odxlinks(odxlinks)
        snref_ctx = SnRefContext()
        req_param1._resolve_snrefs(snref_ctx)
        req_param2._resolve_snrefs(snref_ctx)

        expected_coded_message = bytes([0x12, 0x34])
        expected_param_dict = {
            "SID": 0x12,
            "min_max_parameter": {
                "min_max_parameter": bytes([0x34])
            },
        }

        actual_param_dict = request.decode(expected_coded_message)
        self.assertEqual(dict(actual_param_dict), expected_param_dict)

        actual_coded_message = request.encode(**expected_param_dict)  # type: ignore[arg-type]
        self.assertEqual(actual_coded_message, expected_coded_message)

    def test_physical_constant_parameter(self) -> None:
        odxlinks = OdxLinkDatabase()
        diag_coded_type = StandardLengthType(
            base_data_type=DataType.A_UINT32,
            bit_length=8,
        )
        offset = 0x34
        dop = DataObjectProperty(
            odx_id=OdxLinkId("DOP_ID", doc_frags),
            short_name="DOP",
            diag_coded_type=diag_coded_type,
            physical_type=PhysicalType(base_data_type=DataType.A_INT32),
            compu_method=LinearCompuMethod(
                category=CompuCategory.LINEAR,
                compu_internal_to_phys=CompuInternalToPhys(compu_scales=[
                    CompuScale(
                        compu_rational_coeffs=CompuRationalCoeffs(
                            value_type=DataType.A_INT32,
                            numerators=[offset, 1],
                            denominators=[1],
                        ),
                        domain_type=DataType.A_INT32,
                        range_type=DataType.A_INT32),
                ]),
                internal_type=DataType.A_INT32,
                physical_type=DataType.A_INT32,
            ),
        )
        odxlinks.update(dop._build_odxlinks())
        req_param1 = CodedConstParameter(
            short_name="SID",
            diag_coded_type=diag_coded_type,
            coded_value_raw=str(0x12),
            byte_position=0,
        )
        req_param2 = PhysicalConstantParameter(
            short_name="physical_constant_parameter",
            physical_constant_value_raw=f"{offset}",
            dop_ref=OdxLinkRef.from_id(dop.odx_id),
        )
        request = Request(
            odx_id=OdxLinkId("request", doc_frags),
            short_name="Request",
            parameters=NamedItemList([req_param1, req_param2]),
        )

        req_param1._resolve_odxlinks(odxlinks)
        req_param2._resolve_odxlinks(odxlinks)

        snref_ctx = SnRefContext()
        req_param1._resolve_snrefs(snref_ctx)
        req_param2._resolve_snrefs(snref_ctx)

        expected_coded_message = bytes([0x12, 0x0])
        expected_param_dict = {"SID": 0x12, "physical_constant_parameter": offset}

        actual_param_dict = request.decode(expected_coded_message)
        self.assertEqual(dict(actual_param_dict), expected_param_dict)

        actual_coded_message = request.encode(**expected_param_dict)
        self.assertEqual(actual_coded_message, expected_coded_message)

        self.assertRaises(DecodeError, request.decode, bytes([0x12, 0x34]))


if __name__ == "__main__":
    unittest.main()



# SPDX-License-Identifier: MIT
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Optional
from xml.etree import ElementTree

from typing_extensions import final, override

from ..decodestate import DecodeState
from ..encodestate import EncodeState
from ..exceptions import DecodeError, EncodeError, odxraise, odxrequire
from ..odxdoccontext import OdxDocContext
from ..odxlink import OdxLinkDatabase, OdxLinkId, OdxLinkRef, resolve_snref
from ..odxtypes import ParameterValue
from ..snrefcontext import SnRefContext
from ..utils import dataclass_fields_asdict
from .parameter import Parameter, ParameterType

if TYPE_CHECKING:
    from ..table import Table
    from ..tablerow import TableRow


@dataclass(kw_only=True)
class TableKeyParameter(Parameter):

    odx_id: OdxLinkId

    # the spec mandates that exactly one of the two attributes must
    # be non-None
    table_ref: OdxLinkRef | None = None
    table_snref: str | None = None

    # the spec mandates that exactly one of the two attributes must
    # be non-None
    table_row_ref: OdxLinkRef | None = None
    table_row_snref: str | None = None

    @staticmethod
    @override
    def from_et(et_element: ElementTree.Element, context: OdxDocContext) -> "TableKeyParameter":
        kwargs = dataclass_fields_asdict(Parameter.from_et(et_element, context))

        odx_id = odxrequire(OdxLinkId.from_et(et_element, context))

        table_ref = OdxLinkRef.from_et(et_element.find("TABLE-REF"), context)
        table_snref = None
        if (table_snref_elem := et_element.find("TABLE-SNREF")) is not None:
            table_snref = odxrequire(table_snref_elem.get("SHORT-NAME"))

        table_row_ref = OdxLinkRef.from_et(et_element.find("TABLE-ROW-REF"), context)
        table_row_snref = None
        if (table_row_snref_elem := et_element.find("TABLE-ROW-SNREF")) is not None:
            table_row_snref = odxrequire(table_row_snref_elem.get("SHORT-NAME"))

        return TableKeyParameter(
            odx_id=odx_id,
            table_ref=table_ref,
            table_snref=table_snref,
            table_row_ref=table_row_ref,
            table_row_snref=table_row_snref,
            **kwargs)

    def __post_init__(self) -> None:
        self._table: Table
        self._table_row: TableRow | None = None

    @property
    @override
    def parameter_type(self) -> ParameterType:
        return "TABLE-KEY"

    @override
    def _build_odxlinks(self) -> dict[OdxLinkId, Any]:
        result = super()._build_odxlinks()

        result[self.odx_id] = self

        return result

    @override
    def _resolve_odxlinks(self, odxlinks: OdxLinkDatabase) -> None:
        super()._resolve_odxlinks(odxlinks)

        # Either table_ref or table_row_ref will be defined
        if self.table_ref is not None:
            if TYPE_CHECKING:
                self._table = odxlinks.resolve(self.table_ref, Table)
            else:
                self._table = odxlinks.resolve(self.table_ref)

        elif self.table_row_ref is not None:
            if TYPE_CHECKING:
                self._table_row = odxlinks.resolve(self.table_row_ref, TableRow)
            else:
                self._table_row = odxlinks.resolve(self.table_row_ref)

            # be aware that we cannot simply use
            # `self._table_row.table` here because the table object
            # might not have resolved its references yet because the
            # order of reference resolution is undefined
            if TYPE_CHECKING:
                self._table = odxlinks.resolve(self._table_row.table_ref, Table)
            else:
                self._table = odxlinks.resolve(self._table_row.table_ref)

    @override
    def _resolve_snrefs(self, context: SnRefContext) -> None:
        super()._resolve_snrefs(context)

        if self.table_snref is not None:
            tables = odxrequire(context.diag_layer).diag_data_dictionary_spec.tables
            if TYPE_CHECKING:
                self._table = resolve_snref(self.table_snref, tables, Table)
            else:
                self._table = resolve_snref(self.table_snref, tables)

        if self.table_row_snref is not None:
            if TYPE_CHECKING:
                self._table_row = resolve_snref(self.table_row_snref, self._table.table_rows,
                                                TableRow)
            else:
                self._table_row = resolve_snref(self.table_row_snref, self._table.table_rows)

    @property
    def table(self) -> "Table":
        return self._table

    @property
    def table_row(self) -> Optional["TableRow"]:
        return self._table_row

    @property
    @override
    def is_required(self) -> bool:
        # TABLE-KEY parameters can be implicitly determined from the
        # corresponding TABLE-STRUCT
        return False

    @property
    @override
    def is_settable(self) -> bool:
        return True

    @override
    @final
    def _encode_positioned_into_pdu(self, physical_value: ParameterValue | None,
                                    encode_state: EncodeState) -> None:
        # if you get this exception, you ought to use
        # `.encode_placeholder_into_pdu()` followed by (after the
        # value of the table key has been determined)
        # `.encode_value_into_pdu()`.
        raise RuntimeError("_encode_positioned_into_pdu() cannot be called for table keys.")

    def encode_placeholder_into_pdu(self, physical_value: ParameterValue | None,
                                    encode_state: EncodeState) -> None:

        if physical_value is not None:
            key_dop = self.table.key_dop
            if key_dop is None:
                odxraise(
                    f"Table '{self.table.short_name}' does not define "
                    f"a KEY-DOP, but is used by TABLE-KEY parameter "
                    f"'{self.short_name}'", EncodeError)
                return

            if not isinstance(physical_value, str):
                odxraise(f"Invalid type for for table key '{self.short_name}' specified. "
                         f"(expect name of table row.)")

            tkv = encode_state.table_keys.get(self.short_name)
            if tkv is not None and tkv != physical_value:
                odxraise(f"Got conflicting values for table key {self.short_name}: "
                         f"{tkv} and {physical_value!r}")

            encode_state.table_keys[self.short_name] = physical_value

        pos = encode_state.cursor_byte_position
        if self.byte_position is not None:
            pos = encode_state.origin_byte_position + self.byte_position
        encode_state.key_pos[self.short_name] = pos
        encode_state.cursor_byte_position = pos
        encode_state.cursor_bit_position = self.bit_position or 0

        key_dop = self.table.key_dop
        if key_dop is None:
            odxraise(f"No KEY-DOP specified for table {self.table.short_name}")
            return

        sz = key_dop.get_static_bit_length()
        if sz is None:
            odxraise("The DOP of table key {self.short_name} must exhibit a fixed size.",
                     EncodeError)
            return

        # emplace a value of zero into the encode state, but pretend the bits not to be used
        n = sz + encode_state.cursor_bit_position
        tmp_val = b'\x00' * ((n + 7) // 8)
        encode_state.emplace_bytes(tmp_val, obj_used_mask=tmp_val)

        encode_state.cursor_bit_position = 0

    def encode_value_into_pdu(self, encode_state: EncodeState) -> None:
        key_dop = self.table.key_dop
        if key_dop is None:
            odxraise(
                f"Table '{self.table.short_name}' does not define "
                f"a KEY-DOP, but is used by TABLE-KEY parameter "
                f"'{self.short_name}'", EncodeError)
            return

        if self.short_name not in encode_state.table_keys:
            odxraise(f"Table key {self.short_name} has not been defined before "
                     f"it is required.", EncodeError)
            return
        else:
            tr_short_name = encode_state.table_keys[self.short_name]

        # We need to encode the table key using the associated DOP into the PDU.
        tr_candidates = [x for x in self.table.table_rows if x.short_name == tr_short_name]
        if len(tr_candidates) == 0:
            odxraise(f"No table row with short name '{tr_short_name}' found", EncodeError)
            return
        elif len(tr_candidates) > 1:
            odxraise(f"Multiple rows exhibiting short name '{tr_short_name}'", EncodeError)
        tr = tr_candidates[0]

        encode_state.cursor_byte_position = encode_state.key_pos[self.short_name]
        encode_state.cursor_bit_position = self.bit_position or 0

        key_dop.encode_into_pdu(encode_state=encode_state, physical_value=odxrequire(tr.key))

    @override
    def _decode_positioned_from_pdu(self, decode_state: DecodeState) -> ParameterValue:
        if self.table_row is not None:
            # the table row to be used is statically specified -> no
            # need to decode anything!
            phys_val = self.table_row.short_name
        else:
            # Use DOP to decode
            key_dop = odxrequire(self.table.key_dop)
            key_dop_val = key_dop.decode_from_pdu(decode_state)

            table_row_candidates = [x for x in self.table.table_rows if x.key == key_dop_val]
            if len(table_row_candidates) == 0:
                raise DecodeError(f"No table row exhibiting the key '{str(key_dop_val)}' found")
            elif len(table_row_candidates) > 1:
                raise DecodeError(
                    f"Multiple rows exhibiting key '{str(key_dop_val)}' found in table")
            table_row = table_row_candidates[0]
            phys_val = table_row.short_name

            # update the decode_state's table key
            decode_state.table_keys[self.short_name] = table_row

        return phys_val



# SPDX-License-Identifier: MIT
from dataclasses import dataclass, field
from typing import Any
from xml.etree import ElementTree

from typing_extensions import override

from .complexdop import ComplexDop
from .compositecodec import (composite_codec_decode_from_pdu, composite_codec_encode_into_pdu,
                             composite_codec_get_free_parameters,
                             composite_codec_get_required_parameters,
                             composite_codec_get_static_bit_length)
from .decodestate import DecodeState
from .encodestate import EncodeState
from .exceptions import DecodeError, odxraise
from .nameditemlist import NamedItemList
from .odxdoccontext import OdxDocContext
from .odxlink import OdxLinkDatabase, OdxLinkId
from .odxtypes import ParameterValue
from .parameters.createanyparameter import create_any_parameter_from_et
from .parameters.parameter import Parameter
from .snrefcontext import SnRefContext
from .utils import dataclass_fields_asdict


@dataclass(kw_only=True)
class BasicStructure(ComplexDop):
    """Base class for structure-like objects

    "Structure-like" objects are structures as well as environment
    data objects. All structure-like objects adhere to the
    `CompositeCodec` type protocol.
    """
    byte_size: int | None = None
    parameters: NamedItemList[Parameter] = field(default_factory=NamedItemList)

    @property
    def required_parameters(self) -> list[Parameter]:
        return composite_codec_get_required_parameters(self)

    @property
    def free_parameters(self) -> list[Parameter]:
        return composite_codec_get_free_parameters(self)

    @staticmethod
    def from_et(et_element: ElementTree.Element, context: OdxDocContext) -> "BasicStructure":
        """Read a BASIC-STRUCTURE."""
        kwargs = dataclass_fields_asdict(ComplexDop.from_et(et_element, context))

        byte_size_str = et_element.findtext("BYTE-SIZE")
        byte_size = int(byte_size_str) if byte_size_str is not None else None
        parameters = NamedItemList([
            create_any_parameter_from_et(et_parameter, context)
            for et_parameter in et_element.iterfind("PARAMS/PARAM")
        ])

        return BasicStructure(byte_size=byte_size, parameters=parameters, **kwargs)

    def _build_odxlinks(self) -> dict[OdxLinkId, Any]:
        result = super()._build_odxlinks()

        for param in self.parameters:
            result.update(param._build_odxlinks())

        return result

    def _resolve_odxlinks(self, odxlinks: OdxLinkDatabase) -> None:
        super()._resolve_odxlinks(odxlinks)

        for param in self.parameters:
            param._resolve_odxlinks(odxlinks)

    def _resolve_snrefs(self, context: SnRefContext) -> None:
        context.parameters = self.parameters

        super()._resolve_snrefs(context)
        for param in self.parameters:
            param._resolve_snrefs(context)

        context.parameters = None

    def get_static_bit_length(self) -> int | None:
        # Explicit size was specified, so we do not need to look at
        # the list of parameters
        if self.byte_size is not None:
            return 8 * self.byte_size

        return composite_codec_get_static_bit_length(self)

    def print_free_parameters_info(self) -> None:
        """Return a human readable description of the structure's
        free parameters.
        """
        from .parameterinfo import parameter_info

        print(parameter_info(self.free_parameters), end="")

    @override
    def encode_into_pdu(self, physical_value: ParameterValue | None,
                        encode_state: EncodeState) -> None:
        orig_pos = encode_state.cursor_byte_position

        composite_codec_encode_into_pdu(self, physical_value, encode_state)

        if self.byte_size is not None:
            actual_len = encode_state.cursor_byte_position - orig_pos

            if actual_len < self.byte_size:
                # Padding bytes are needed. We add an empty object at
                # the position directly after the structure and let
                # EncodeState add the padding as needed.
                encode_state.cursor_byte_position = encode_state.origin_byte_position + self.byte_size
                # Padding bytes needed. these count as "used".
                encode_state.coded_message += b"\x00" * (self.byte_size - actual_len)
                encode_state.used_mask += b"\xff" * (self.byte_size - actual_len)

    @override
    def decode_from_pdu(self, decode_state: DecodeState) -> ParameterValue:
        orig_pos = decode_state.cursor_byte_position

        result = composite_codec_decode_from_pdu(self, decode_state)

        if self.byte_size is not None:
            n = decode_state.cursor_byte_position - orig_pos
            if n > self.byte_size:
                odxraise(
                    f"Attempted to decode too large instance of structure "
                    f"{self.short_name} ({n} instead at most "
                    f"{self.byte_size} bytes)", DecodeError)
                return result

            decode_state.cursor_byte_position = orig_pos + self.byte_size

        return result



# SPDX-License-Identifier: MIT
import warnings
from dataclasses import dataclass
from typing import Any
from xml.etree import ElementTree

from .basecomparam import BaseComparam
from .comparam import Comparam
from .complexcomparam import ComplexComparam, ComplexValue, create_complex_value_from_et
from .description import Description
from .exceptions import OdxWarning, odxraise, odxrequire
from .odxdoccontext import OdxDocContext
from .odxlink import OdxLinkDatabase, OdxLinkId, OdxLinkRef
from .snrefcontext import SnRefContext


@dataclass(kw_only=True)
class ComparamInstance:
    """
    This class represents a communication parameter.

    Be aware that the ODX specification calls this class COMPARAM-REF!
    """
    value: str | ComplexValue
    description: Description | None = None
    protocol_snref: str | None = None
    prot_stack_snref: str | None = None
    spec_ref: OdxLinkRef

    @property
    def spec(self) -> BaseComparam:
        return self._spec

    @property
    def short_name(self) -> str:
        return self.spec.short_name

    @staticmethod
    def from_et(et_element: ElementTree.Element, context: OdxDocContext) -> "ComparamInstance":
        spec_ref = odxrequire(OdxLinkRef.from_et(et_element, context))

        # ODX standard v2.0.0 defined only VALUE. ODX v2.0.1 decided
        # to break things and change it to a choice between SIMPLE-VALUE
        # and COMPLEX-VALUE
        value: str | list[str | ComplexValue]
        if et_element.find("VALUE") is not None:
            value = odxrequire(et_element.findtext("VALUE"))
        elif et_element.find("SIMPLE-VALUE") is not None:
            value = odxrequire(et_element.findtext("SIMPLE-VALUE"))
        else:
            value = create_complex_value_from_et(odxrequire(et_element.find("COMPLEX-VALUE")))

        description = Description.from_et(et_element.find("DESC"), context)

        protocol_snref = None
        if (psnref_elem := et_element.find("PROTOCOL-SNREF")) is not None:
            protocol_snref = psnref_elem.get("SHORT-NAME")

        prot_stack_snref = None
        if (psnref_elem := et_element.find("PROT-STACK-SNREF")) is not None:
            prot_stack_snref = psnref_elem.get("SHORT-NAME")

        return ComparamInstance(
            value=value,
            spec_ref=spec_ref,
            description=description,
            protocol_snref=protocol_snref,
            prot_stack_snref=prot_stack_snref,
        )

    def _build_odxlinks(self) -> dict[OdxLinkId, Any]:
        return {}

    def _resolve_odxlinks(self, odxlinks: OdxLinkDatabase) -> None:
        self._spec = odxlinks.resolve(self.spec_ref, BaseComparam)

    def _resolve_snrefs(self, context: SnRefContext) -> None:
        pass

    def get_value(self) -> str:
        """Retrieve the value of a simple communication parameter

        This takes the default value of the comparam (if any) into
        account.
        """

        if not isinstance(self.spec, Comparam):
            odxraise()

        result: Any = None
        if self.value:
            result = self.value
        else:
            result = self.spec.physical_default_value

        if not isinstance(result, str):
            odxraise()

        return result

    def get_subvalue(self, subparam_name: str) -> str | None:
        """Retrieve the value of a complex communication parameter's sub-parameter by name

        This takes the default value of the comparam (if any) into
        account.
        """
        comparam_spec = self.spec
        if not isinstance(comparam_spec, ComplexComparam):
            odxraise()

        value_list = self.value
        if not isinstance(value_list, list):
            warnings.warn(
                f"The values of complex communication parameter "
                f"'{self.short_name}' are not specified "
                f"correctly.",
                OdxWarning,
                stacklevel=1,
            )
            return None

        name_list = [cp.short_name for cp in comparam_spec.subparams]
        try:
            idx = name_list.index(subparam_name)
            subparam = comparam_spec.subparams[idx]
        except ValueError:
            warnings.warn(
                f"Communication parameter '{self.short_name}' "
                f"does not specify a '{subparam_name}' sub-parameter.",
                OdxWarning,
                stacklevel=1,
            )
            return None

        result = value_list[idx]
        if result is None and isinstance(subparam, (Comparam, ComplexComparam)):
            result = subparam.physical_default_value
        if not isinstance(result, str):
            odxraise()

        return result



# SPDX-License-Identifier: MIT
from dataclasses import dataclass, field
from typing import Any, cast
from xml.etree import ElementTree

from ..additionalaudience import AdditionalAudience
from ..admindata import AdminData
from ..companydata import CompanyData
from ..diagcomm import DiagComm
from ..diagdatadictionaryspec import DiagDataDictionarySpec
from ..diagservice import DiagService
from ..element import IdentifiableElement
from ..exceptions import odxassert, odxraise, odxrequire
from ..functionalclass import FunctionalClass
from ..library import Library
from ..nameditemlist import NamedItemList
from ..odxdoccontext import OdxDocContext
from ..odxlink import DocType, OdxDocFragment, OdxLinkDatabase, OdxLinkId, OdxLinkRef
from ..request import Request
from ..response import Response
from ..singleecujob import SingleEcuJob
from ..snrefcontext import SnRefContext
from ..specialdatagroup import SpecialDataGroup
from ..statechart import StateChart
from ..subcomponent import SubComponent
from ..utils import dataclass_fields_asdict
from .diaglayertype import DiagLayerType


@dataclass(kw_only=True)
class DiagLayerRaw(IdentifiableElement):
    """This class internalizes all data represented by the DIAG-LAYER
    XML tag and its derivatives.

    It does *not* deal with value inheritance.
    """

    variant_type: DiagLayerType
    admin_data: AdminData | None = None
    company_datas: NamedItemList[CompanyData] = field(default_factory=NamedItemList)
    functional_classes: NamedItemList[FunctionalClass] = field(default_factory=NamedItemList)
    diag_data_dictionary_spec: DiagDataDictionarySpec | None = None
    diag_comms_raw: list[OdxLinkRef | DiagComm] = field(default_factory=list)
    requests: NamedItemList[Request] = field(default_factory=NamedItemList)
    positive_responses: NamedItemList[Response] = field(default_factory=NamedItemList)
    negative_responses: NamedItemList[Response] = field(default_factory=NamedItemList)
    global_negative_responses: NamedItemList[Response] = field(default_factory=NamedItemList)
    import_refs: list[OdxLinkRef] = field(default_factory=list)
    state_charts: NamedItemList[StateChart] = field(default_factory=NamedItemList)
    additional_audiences: NamedItemList[AdditionalAudience] = field(default_factory=NamedItemList)
    sub_components: NamedItemList[SubComponent] = field(default_factory=NamedItemList)
    libraries: NamedItemList[Library] = field(default_factory=NamedItemList)
    sdgs: list[SpecialDataGroup] = field(default_factory=list)

    @property
    def diag_comms(self) -> NamedItemList[DiagComm]:
        return self._diag_comms

    @property
    def diag_services(self) -> NamedItemList[DiagService]:
        return self._diag_services

    @property
    def services(self) -> NamedItemList[DiagService]:
        """This property is an alias for `.diag_services`"""
        return self._diag_services

    @property
    def single_ecu_jobs(self) -> NamedItemList[SingleEcuJob]:
        return self._single_ecu_jobs

    @staticmethod
    def from_et(et_element: ElementTree.Element, context: OdxDocContext) -> "DiagLayerRaw":
        try:
            variant_type = DiagLayerType(et_element.tag)
        except ValueError:
            variant_type = cast(DiagLayerType, None)
            odxraise(f"Encountered unknown diagnostic layer type '{et_element.tag}'")

        short_name = odxrequire(et_element.findtext("SHORT-NAME"))

        # extend the applicable ODX "document fragments" for the diag layer objects
        context.doc_fragments.append(OdxDocFragment(short_name, DocType.LAYER))
        kwargs = dataclass_fields_asdict(IdentifiableElement.from_et(et_element, context))

        admin_data = None
        if (admin_data_elem := et_element.find("ADMIN-DATA")) is not None:
            admin_data = AdminData.from_et(admin_data_elem, context)

        company_datas = NamedItemList([
            CompanyData.from_et(cd_el, context)
            for cd_el in et_element.iterfind("COMPANY-DATAS/COMPANY-DATA")
        ])

        functional_classes = NamedItemList([
            FunctionalClass.from_et(fc_el, context)
            for fc_el in et_element.iterfind("FUNCT-CLASSS/FUNCT-CLASS")
        ])

        diag_data_dictionary_spec = None
        if (ddds_elem := et_element.find("DIAG-DATA-DICTIONARY-SPEC")) is not None:
            diag_data_dictionary_spec = DiagDataDictionarySpec.from_et(ddds_elem, context)

        diag_comms_raw: list[OdxLinkRef | DiagComm] = []
        if (dc_elems := et_element.find("DIAG-COMMS")) is not None:
            for dc_proxy_elem in dc_elems:
                dc: OdxLinkRef | DiagComm
                if dc_proxy_elem.tag == "DIAG-COMM-REF":
                    dc = OdxLinkRef.from_et(dc_proxy_elem, context)
                elif dc_proxy_elem.tag == "DIAG-SERVICE":
                    dc = DiagService.from_et(dc_proxy_elem, context)
                else:
                    odxassert(dc_proxy_elem.tag == "SINGLE-ECU-JOB")
                    dc = SingleEcuJob.from_et(dc_proxy_elem, context)

                diag_comms_raw.append(dc)

        requests = NamedItemList([
            Request.from_et(rq_elem, context) for rq_elem in et_element.iterfind("REQUESTS/REQUEST")
        ])

        positive_responses = NamedItemList([
            Response.from_et(rs_elem, context)
            for rs_elem in et_element.iterfind("POS-RESPONSES/POS-RESPONSE")
        ])

        negative_responses = NamedItemList([
            Response.from_et(rs_elem, context)
            for rs_elem in et_element.iterfind("NEG-RESPONSES/NEG-RESPONSE")
        ])

        global_negative_responses = NamedItemList([
            Response.from_et(rs_elem, context)
            for rs_elem in et_element.iterfind("GLOBAL-NEG-RESPONSES/GLOBAL-NEG-RESPONSE")
        ])

        import_refs = [
            OdxLinkRef.from_et(el, context) for el in et_element.iterfind("IMPORT-REFS/IMPORT-REF")
        ]

        state_charts = NamedItemList([
            StateChart.from_et(el, context)
            for el in et_element.iterfind("STATE-CHARTS/STATE-CHART")
        ])

        additional_audiences = NamedItemList([
            AdditionalAudience.from_et(el, context)
            for el in et_element.iterfind("ADDITIONAL-AUDIENCES/ADDITIONAL-AUDIENCE")
        ])

        sub_components = NamedItemList([
            SubComponent.from_et(el, context)
            for el in et_element.iterfind("SUB-COMPONENTS/SUB-COMPONENT")
        ])

        libraries = NamedItemList(
            [Library.from_et(el, context) for el in et_element.iterfind("LIBRARYS/LIBRARY")])

        sdgs = [SpecialDataGroup.from_et(sdge, context) for sdge in et_element.iterfind("SDGS/SDG")]

        # Create DiagLayer
        return DiagLayerRaw(
            variant_type=variant_type,
            admin_data=admin_data,
            company_datas=company_datas,
            functional_classes=functional_classes,
            diag_data_dictionary_spec=diag_data_dictionary_spec,
            diag_comms_raw=diag_comms_raw,
            requests=requests,
            positive_responses=positive_responses,
            negative_responses=negative_responses,
            global_negative_responses=global_negative_responses,
            import_refs=import_refs,
            state_charts=state_charts,
            additional_audiences=additional_audiences,
            sub_components=sub_components,
            libraries=libraries,
            sdgs=sdgs,
            **kwargs)

    def _build_odxlinks(self) -> dict[OdxLinkId, Any]:
        """Construct a mapping from IDs to all objects that are contained in this diagnostic layer."""
        odxlinks = {self.odx_id: self}

        if self.admin_data is not None:
            odxlinks.update(self.admin_data._build_odxlinks())
        if self.diag_data_dictionary_spec is not None:
            odxlinks.update(self.diag_data_dictionary_spec._build_odxlinks())

        for company_data in self.company_datas:
            odxlinks.update(company_data._build_odxlinks())
        for functional_class in self.functional_classes:
            odxlinks.update(functional_class._build_odxlinks())
        for dc_proxy in self.diag_comms_raw:
            if isinstance(dc_proxy, OdxLinkRef):
                continue
            odxlinks.update(dc_proxy._build_odxlinks())
        for request in self.requests:
            odxlinks.update(request._build_odxlinks())
        for positive_response in self.positive_responses:
            odxlinks.update(positive_response._build_odxlinks())
        for negative_response in self.negative_responses:
            odxlinks.update(negative_response._build_odxlinks())
        for global_negative_response in self.global_negative_responses:
            odxlinks.update(global_negative_response._build_odxlinks())
        for state_chart in self.state_charts:
            odxlinks.update(state_chart._build_odxlinks())
        for additional_audience in self.additional_audiences:
            odxlinks.update(additional_audience._build_odxlinks())
        for sub_comp in self.sub_components:
            odxlinks.update(sub_comp._build_odxlinks())
        for library in self.libraries:
            odxlinks.update(library._build_odxlinks())
        for sdg in self.sdgs:
            odxlinks.update(sdg._build_odxlinks())

        return odxlinks

    def _resolve_odxlinks(self, odxlinks: OdxLinkDatabase) -> None:
        """Recursively resolve all references."""

        # do ODXLINK reference resolution
        if self.admin_data is not None:
            self.admin_data._resolve_odxlinks(odxlinks)
        if self.diag_data_dictionary_spec is not None:
            self.diag_data_dictionary_spec._resolve_odxlinks(odxlinks)

        for company_data in self.company_datas:
            company_data._resolve_odxlinks(odxlinks)
        for functional_class in self.functional_classes:
            functional_class._resolve_odxlinks(odxlinks)

        # resolve references to diagnostic communication objects and
        # separate them into services and single-ecu jobs
        self._diag_comms = NamedItemList[DiagComm]()
        self._diag_services = NamedItemList[DiagService]()
        self._single_ecu_jobs = NamedItemList[SingleEcuJob]()
        for dc_proxy in self.diag_comms_raw:
            if isinstance(dc_proxy, OdxLinkRef):
                dc = odxlinks.resolve(dc_proxy, DiagComm)
            else:
                dc = dc_proxy
                dc._resolve_odxlinks(odxlinks)

            self._diag_comms.append(dc)

            if isinstance(dc, DiagService):
                self._diag_services.append(dc)
            elif isinstance(dc, SingleEcuJob):
                self._single_ecu_jobs.append(dc)
            else:
                odxraise()

        for request in self.requests:
            request._resolve_odxlinks(odxlinks)
        for positive_response in self.positive_responses:
            positive_response._resolve_odxlinks(odxlinks)
        for negative_response in self.negative_responses:
            negative_response._resolve_odxlinks(odxlinks)
        for global_negative_response in self.global_negative_responses:
            global_negative_response._resolve_odxlinks(odxlinks)
        for state_chart in self.state_charts:
            state_chart._resolve_odxlinks(odxlinks)
        for additional_audience in self.additional_audiences:
            additional_audience._resolve_odxlinks(odxlinks)
        for sub_component in self.sub_components:
            sub_component._resolve_odxlinks(odxlinks)
        for library in self.libraries:
            library._resolve_odxlinks(odxlinks)
        for sdg in self.sdgs:
            sdg._resolve_odxlinks(odxlinks)

    def _resolve_snrefs(self, context: SnRefContext) -> None:
        # do short-name reference resolution
        if self.admin_data is not None:
            self.admin_data._resolve_snrefs(context)
        if self.diag_data_dictionary_spec is not None:
            self.diag_data_dictionary_spec._resolve_snrefs(context)

        for company_data in self.company_datas:
            company_data._resolve_snrefs(context)
        for functional_class in self.functional_classes:
            functional_class._resolve_snrefs(context)
        for dc_proxy in self.diag_comms_raw:
            if isinstance(dc_proxy, OdxLinkRef):
                continue
            dc_proxy._resolve_snrefs(context)
        for request in self.requests:
            request._resolve_snrefs(context)
        for positive_response in self.positive_responses:
            positive_response._resolve_snrefs(context)
        for negative_response in self.negative_responses:
            negative_response._resolve_snrefs(context)
        for global_negative_response in self.global_negative_responses:
            global_negative_response._resolve_snrefs(context)
        for state_chart in self.state_charts:
            state_chart._resolve_snrefs(context)
        for additional_audience in self.additional_audiences:
            additional_audience._resolve_snrefs(context)
        for sub_component in self.sub_components:
            sub_component._resolve_snrefs(context)
        for library in self.libraries:
            library._resolve_snrefs(context)
        for sdg in self.sdgs:
            sdg._resolve_snrefs(context)



# SPDX-License-Identifier: MIT
from dataclasses import dataclass
from typing import Any
from xml.etree import ElementTree

from .odxdoccontext import OdxDocContext
from .odxlink import OdxLinkDatabase, OdxLinkId
from .snrefcontext import SnRefContext


@dataclass(kw_only=True)
class SpecialData:
    """This corresponds to the SD XML tag"""
    semantic_info: str | None = None  # the "SI" attribute
    text_identifier: str | None = None  # the "TI" attribute, specifies the language used
    value: str

    @staticmethod
    def from_et(et_element: ElementTree.Element, context: OdxDocContext) -> "SpecialData":
        semantic_info = et_element.get("SI")
        text_identifier = et_element.get("TI")
        value = et_element.text or ""

        return SpecialData(
            semantic_info=semantic_info, text_identifier=text_identifier, value=value)

    def _build_odxlinks(self) -> dict[OdxLinkId, Any]:
        return {}

    def _resolve_odxlinks(self, odxlinks: OdxLinkDatabase) -> None:
        pass

    def _resolve_snrefs(self, context: SnRefContext) -> None:
        pass



# SPDX-License-Identifier: MIT
import unittest

from odxtools.loadfile import load_pdx_file
from odxtools.statemachine import StateMachine

odxdb = load_pdx_file("./examples/somersault.pdx")


class TestStateChart(unittest.TestCase):

    def test_state_chart(self) -> None:
        ecu = odxdb.ecu_variants.somersault_lazy

        self.assertEqual(len(ecu.state_charts), 2)
        self.assertEqual(ecu.state_charts[0].semantic, "annoyed")
        self.assertEqual(ecu.state_charts[1].semantic, "angry")

        annoyed_chart = ecu.state_charts[0]
        self.assertEqual(len(annoyed_chart.states), 4)
        self.assertEqual(len(annoyed_chart.state_transitions), 6)
        self.assertEqual(annoyed_chart.start_state.short_name, "in_bed")

    def test_state_machine(self) -> None:
        ecu = odxdb.ecu_variants.somersault_lazy
        annoyed_chart = ecu.state_charts.annoyed_chart
        angry_chart = ecu.state_charts.angry_chart

        fsm_angry = StateMachine(ecu, angry_chart)
        fsm_annoyed = StateMachine(ecu, annoyed_chart)

        # the initial states of the state machines are the start
        # states of their respective chart
        self.assertEqual(fsm_angry.active_state, angry_chart.start_state)
        self.assertEqual(fsm_annoyed.active_state, annoyed_chart.start_state)

        # not giving a bribe will fail with the lazy ECU variant if it
        # is in an angry mood
        service = ecu.services.session_start
        for _ in (executor := fsm_angry.execute(service)):
            # the state machine is not supposed to do anything because
            # the precondition to execute the "session_start"
            # service's request is not fulfilled
            self.assertTrue(False)
        self.assertEqual(fsm_angry.active_state, angry_chart.start_state)
        self.assertEqual(fsm_angry.succeeded, False)

        # explicitly not giving a bribe fails as well
        service = ecu.services.session_start
        for _ in (executor := fsm_angry.execute(service, bribe=0)):
            # the state machine is not supposed to do anything because
            # the precondition to execute the "session_start"
            # service's request is not fulfilled
            self.assertTrue(False)
        self.assertEqual(fsm_angry.active_state, angry_chart.start_state)
        self.assertEqual(fsm_angry.succeeded, False)

        # giving a small bribe will get an angry lazy ECU out of bed
        n = 0
        for raw_request in (executor := fsm_angry.execute(ecu.services.session_start, bribe=1)):
            n += 1
            self.assertEqual(raw_request, b'\x10\x00\x01')
            resp = service.positive_responses.session
            raw_response = resp.encode(coded_request=raw_request, can_do_backward_flips="false")
            executor.send(raw_response)
        self.assertEqual(n, 1)
        self.assertEqual(fsm_angry.active_state, angry_chart.states.on_street)
        self.assertEqual(fsm_angry.succeeded, True)

        # if the ECU is just annoyed, it will get out of bed even if
        # no bribe is given. if the ECU responds that it can to
        # backward flips, it will end up in the park, else it will go
        # to the street.
        service = ecu.services.session_start
        for raw_request in (executor := fsm_annoyed.execute(ecu.services.session_start)):
            self.assertEqual(raw_request, b'\x10\x00\x00')
            resp = service.positive_responses.session
            raw_response = resp.encode(coded_request=raw_request, can_do_backward_flips="false")
            executor.send(raw_response)
        self.assertEqual(fsm_annoyed.active_state, annoyed_chart.states.on_street)
        self.assertEqual(fsm_annoyed.succeeded, True)

        service = ecu.services.session_start
        fsm_annoyed.active_state = annoyed_chart.start_state
        for raw_request in (executor := fsm_annoyed.execute(ecu.services.session_start, bribe=0)):
            self.assertEqual(raw_request, b'\x10\x00\x00')
            resp = service.positive_responses.session
            raw_response = resp.encode(coded_request=raw_request, can_do_backward_flips="true")
            executor.send(raw_response)
        self.assertEqual(fsm_annoyed.active_state, annoyed_chart.states.in_park)
        self.assertEqual(fsm_annoyed.succeeded, True)


if __name__ == "__main__":
    unittest.main()
